from construct.core import *
from construct.lib import *

## Extensions are added here
## It may be discussed whether LURK is an extension or not.
## we considered it as an extension to make sure we scale to
## additional extension as well as for readability.
## This is not fully achieved as LURKCapabilitiesResponse
## is defined here as it needs the knowledge of all existing
## extensions.
from pylurk.lurk.struct_lurk_lurk import LURKVersion, LURKType, LURKStatus, ErrorPayload, EmptyPayload
from pylurk.tls13.struct_lurk_tls13 import TLS13Version, TLS13Type, TLS13Status, TLS13Payload


## list the supported extensions
## can parse any value but cannot build any value.
Designation = Enum( BytesInteger( 1 ),
  lurk = 0,
  tls12 = 1,
  tls13 = 2
)


## building generic variables across extensions
## The format of a packet may take any value
## The only valid case is that these unexpected
## values are associated to an ErrorResponse
##
## default is set to Byte so any value can be returned.
Version = Switch( this.designation,
  { 'lurk' : LURKVersion,
    'tls13' : TLS13Version
  }, default=Byte )

Type = Switch( this.designation,
  { 'lurk' : LURKType,
    'tls13' : TLS13Type,
  }, default=Byte)

## The default is set to LURKStatus so an error with a
## bad designation can be returned with a effective error message.
## when the parsing occurs, number is returned.
Status = Switch( this.designation,
     { 'lurk' : LURKStatus,
       'tls13' : TLS13Status,
     }, default=LURKStatus )


## LURK Capability needs global information - that is not
## restricted to the lurk lurk extension:
############# LURKCapabilitiesResponse

Extension = Struct(
  'designation' / Designation,
  'version' / Version
)

LURKSupportedExtensions = Prefixed(
  BytesInteger(2),
  GreedyRange(Extension)
)

LURKCapabilitiesResponse = Struct( \
  'supported_extensions' / LURKSupportedExtensions, \
  'lurk_state' / Bytes(4)
)

## The Payload lists the acceptable payload
## and raises an Error otherwise.
## The server will be responsible to return the
## appropriated ErrorResponse from that Error.
LURKPayload = Switch(this._type,
  { 'capabilities' : Switch( this._status,
      { 'request' : EmptyPayload,
        'success' : LURKCapabilitiesResponse
      }, default=ErrorPayload
    ),
    'ping' : EmptyPayload,
  }, default=ErrorPayload
)

### generic structure for LURK Messages
LURKHeader = Struct (
  '_name' / Computed('LURK message'),
  'designation' / Designation,
  'version' / Version,
  'type' / Type,
  'status' / Status,
  'id' / BytesInteger(8),
)

LURKMessage = Struct (
  '_name' / Computed('LURK message'),
  'designation' / Designation,
  'version' / Version,
  'type' / Type,
  'status' / Status,
  'id' / BytesInteger(8),
  ## these values are used by Payloads
  '_type' / Computed(this.type),
  '_status' / Computed(this.status),
  ## default parameters for TLS13
  '_certificate_type' / Computed( 'X509' ),
  'payload' / Prefixed( BytesInteger(4),
#     IfThenElse( this.status in [ 'request', 'success' ],
       Switch( this.designation,
         { 'lurk' : LURKPayload,
           'tls13' : TLS13Payload,
         }, default=ErrorPayload ), # other designation are
                                     # associated to ErrorResponse
#       ErrorPayload) # other status are associated to ErrorResponse
  ) # end of prefix
)

LURKMessageList = GreedyRange( LURKMessage )




#from secrets import randbits, token_bytes
#from  os.path import join
#from copy import deepcopy
from typing import TypeVar
#Union, NoReturn, List
from cryptography.hazmat.primitives.hashes import Hash
#, SHA256, SHA384, SHA512
from cryptography.hazmat.primitives.kdf.hkdf import HKDF, HKDFExpand
#from cryptography.hazmat.primitives.hmac import HMAC
#from construct.core import *
#from construct.lib import *

import pylurk.debug
import pylurk.tls13.struct_tls13 as lurk
from pylurk.lurk.lurk_lurk import ImplementationError
#import pylurk.debug
#from pylurk.debug import get_struct, get_struct_index
#from pylurk.tls13.crypto_suites import SigScheme, CipherSuite

TlsHandshake = TypeVar( 'TlsHandshake' )

class KeyScheduler:

  def __init__( self, tls_hash, shared_secret:bytes=None, 
                psk:bytes=None, is_ext=False, debug=None ):
    self.secrets = { 'b' : None, 'e_s' : None, 'e_x' : None,\
                    'h_c' : None, 'h_s' : None, 'a_c' : None,\
                    'a_s' : None, 'x' : None, 'r' : None }
    self.tickets = []
    self.psk = psk
    self.is_ext = is_ext
    self.tls_hash = pylurk.tls13.crypto_suites.hash_sanity_check( tls_hash )

    self.shared_secret = shared_secret
    self.debug = debug

    ## initializaton of secret
    self.early_secret = None
    self.handshake_secret = None
    self.master_secret = None

  def hkdf_expand_label( self, secret,\
                         label, \
                         context, \
                         length ):
    hkdf_label = lurk.HkdfLabel.build( \
      { 'length' : length,
        'label' : b"tls13 " + label, \
        'context' : context } )
    return HKDFExpand( algorithm=self.tls_hash,\
                       length=length,\
                       info=hkdf_label ).derive( secret )

  def derive_secret( self, secret, label, transcript_h ):
    """ derive secret function

    Args:
      secret: the secret
      label: the label
      transcript_h : the transcript hash associated to the messages.
        Note that this is a main difference with Derive-Secret as
        described in RFC4664. Derive-Secret takes the messages and
        performs the Transcript-Hash of the messages.
    """
    return self.hkdf_expand_label( secret=secret, \
                                   label=label, \
                                   context=transcript_h, \
                                   length=self.tls_hash.digest_size )
#                                   algorithm=self.tls_hash )

  def process_early_secret( self ):

    if self.psk is None:
      self.psk = b"\x00" * self.tls_hash.digest_size

    self.early_secret = HKDF(
            algorithm=self.tls_hash,
            length=self.tls_hash.digest_size,
            info=None, # None or \x00 zero length string
            salt=b"\x00",
        )._extract( self.psk )

    if self.debug is not None:
      self.debug.handle_bin( "psk", self.psk )
      self.debug.handle_bin( "early_secret",  self.early_secret )


  def empty_transcript_h( self ):
    h = Hash( self.tls_hash )
    h.update( b'' )
    return h.finalize()


  def process_handshake_and_master_secret( self, shared_secret=None ):

    if shared_secret is not None:
      self.shared_secret = shared_secret
    empty_transcript = self.empty_transcript_h( )

    if self.shared_secret is None:
      self.shared_secret = b"\x00" * self.tls_hash.digest_size

    handshake_secret_salt = self.derive_secret(\
      self.early_secret,\
      b'derived',\
      empty_transcript )

    self.handshake_secret = HKDF(
      algorithm=self.tls_hash,
      salt=handshake_secret_salt,
      info=None,
#      backend=backend,
      length=self.tls_hash.digest_size )._extract( self.shared_secret )

    master_secret_salt = self.derive_secret(\
      self.handshake_secret,\
      b'derived',\
      empty_transcript )

    self.master_secret = HKDF(
      algorithm=self.tls_hash,
      salt=master_secret_salt,
      info=None,
      length=self.tls_hash.digest_size )._extract( b"\x00" * self.tls_hash.digest_size )
    if self.debug is not None:
      self.debug.handle_bin( "empty_transcript", empty_transcript )
      self.debug.handle_bin( "handshake_secret", self.handshake_secret )
      self.debug.handle_bin( "master_secret", self.master_secret )


  def process( self, secret_list:list, handshake:TlsHandshake ) -> None:

    ## early secrets
    if 'b' in secret_list or 'e_s' in secret_list or  'e_x' in secret_list:
      if self.early_secret is None :
        self.process_early_secret( )
      s = self.early_secret
      ## generating traffic secrets are derived
      if 'b' in secret_list :
        t = self.empty_transcript_h()
        if self.is_ext is True:
          label = b'ext binder'
        else:
          label = b'res binder'
        self.secrets[ 'b' ] = self.derive_secret( s, label, t )
        if self.debug is not None:
          self.debug.handle_bin( "transcript h ['b']", t )
          self.debug.handle_bin( "binder_key", self.secrets[ 'b' ] )
      if 'e_s' in secret_list or 'e_x' in secret_list:
        t = handshake.transcript_hash( 'e' )
        if self.debug is not None:
          self.debug.handle_bin( "transcript h ['e_s', 'e_x']", t )
        if 'e_s' in secret_list:
          self.secrets[ 'e_s' ] = self.derive_secret( s, b'c e traffic', t )
          if self.debug is not None:
            self.debug.handle_bin( "client_early_traffic_secret", self.secrets[ 'e_s' ] )
        if 'e_x' in secret_list:
          self.secrets[ 'e_x' ] = self.derive_secret( s, b'e exp master', t )
          if self.debug is not None:
            self.debug.handle_bin( "early_exporter_master_secret", self.secrets[ 'e_x' ] )
    # handshake and other secrets
    if 'h_c' in secret_list or  'h_s' in secret_list or \
       'a_c' in secret_list or  'a_s' in secret_list or \
       'r' in secret_list:
      if self.early_secret is None :
        self.process_early_secret( )
      if self.handshake_secret is None:
        self.process_handshake_and_master_secret( )

      if 'h_c' in secret_list or  'h_s' in secret_list:
        t = handshake.transcript_hash( 'h' )
        s = self.handshake_secret
        self.secrets[ 'h_c' ] = self.derive_secret( s, b'c hs traffic', t )
        self.secrets[ 'h_s' ] = self.derive_secret( s, b's hs traffic', t )
        if self.debug is not None:
          self.debug.handle_bin( "transcript h ['h_c', 'h_s']", t )
          self.debug.handle_bin( "client_handshake_traffic_secret", self.secrets[ 'h_c' ] )
          self.debug.handle_bin( "server_handshake_traffic_secret", self.secrets[ 'h_s' ] )

      if 'a_c' in secret_list or  'a_s' in secret_list or 'x' in secret_list:
        t = handshake.transcript_hash( 'a' )
        s = self.master_secret
        self.secrets[ 'a_c' ] = self.derive_secret( s, b'c ap traffic', t )
        self.secrets[ 'a_s' ] = self.derive_secret( s, b's ap traffic', t )
        if self.debug is not None:
          self.debug.handle_bin( "transcript h ['a_c', 'a_s']", t )
          self.debug.handle_bin( "client_application_traffic_secret_0", self.secrets[ 'a_c' ] )
          self.debug.handle_bin( "server_application_traffic_secret_0", self.secrets[ 'a_s' ] )
        if 'x' in secret_list:
          self.secrets[ 'x' ] = self.derive_secret( s, b'exp master', t )
          if self.debug is not None:
            self.debug.handle_bin( "exporter_master_secret", self.secrets[ 'x' ] )

      if 'r' in secret_list:
        t = handshake.transcript_hash( 'r' )
        s = self.master_secret
        self.secrets[ 'r' ] = self.derive_secret( s, b'res master', t )
        if self.debug is not None:
          self.debug.handle_bin( "transcript h ['r']", t )
          self.debug.handle_bin( "resumption_master_secret", self.secrets[ 'r' ] )

  def finished_key( self, role ):
    if role == 'server' :
      secret = self.secrets[ 'h_s' ]
    elif role == 'client' :
      secret = self.secrets[ 'h_c' ]
    elif role == 'binder' :
      secret = self.secrets[ 'b' ]
    else:
      raise ImplementationError( f"unknown role {role}" )

    return self.hkdf_expand_label( secret=secret,
                     label=b'finished', context=b'',\
                     length=self.tls_hash.digest_size )

  def compute_psk( self, ticket_nonce ) -> bytes: ## or None
    return self.hkdf_expand_label( secret=self.secrets[ 'r' ],\
                         label=b"resumption", \
                         context=ticket_nonce, \
                         length=self.tls_hash.digest_size )
#                         backend=default_backend() ) #, \

  def next_generation_application_traffic_secret( self, secret ):
    return self.hkdf_expand_label( secret=secret, label=b'traffic upd', \
             context=b'', length=self.tls_hash.digest_size )



from test_utils import * 
from secrets import randbelow
from copy import deepcopy

## inserting the src in the search path 
import sys
sys.path.insert( 0, '../src/')

from pylurk.struct_lurk import *

## Ability to receive unexpected header
print( Designation.parse( b'\x05' ) )
print( Designation.build( 5 )  )

## ability to parse
print( Version.parse( b'\x00', designation='lurk' ) ) ## expected
print( Version.parse( b'\x06', designation='lurk' ) ) ## unexpected version
print( Version.parse( b'\x06', designation=b'\x05' ) ) ## unexpected version / designation


print( Version.build( 6, designation=5 ) ) ## unexpected version / designation

ping_req = { 'designation' : 'lurk', 
             'version' : 'v1', 
             'type'  : 'ping',
             'status' : 'request',
             'id' : randbelow( 2  ** 64 ),  
             'payload' : {} }

ping_resp = { 'designation' : 'lurk', 
              'version' : 'v1', 
              'type'  : 'ping',
              'status' : 'success',
              'id' : randbelow( 2  ** 64 ),  
              'payload' : {} }

cap_req = { 'designation' : 'lurk', 
            'version' : 'v1', 
            'type'  : 'capabilities',
            'status' : 'request',
            'id' : randbelow( 2  ** 64 ),  
            'payload' : {} }

cap_resp = { 'designation' : 'lurk', 
             'version' : 'v1', 
             'type'  : 'capabilities',
             'status' : 'success',
             'id' : randbelow( 2  ** 64 ),  
             'payload' : { \
                'supported_extensions' : [ { 'designation' : 'lurk', 'version' : 'v1'},
                                            { 'designation' : 'tls13', 'version' : 'v1'} ], 
                 'lurk_state' : b'1234' } }

# this error may be returned an error occurs inside the extension
error_resp = { 'designation' : 'lurk', 
               'version' : 'v1', 
               'type'  : 'capabilities',
               'status' : 'invalid_format',
               'id' : randbelow( 2  ** 64 ),
               'payload' : { 'lurk_state' : b'1234' } }

# these errors indicates there is no extensions, but may
# be sent by the server as well
# undefined values MUST be replaced by integers to allow the 
# response to be built 
error_resp2 = { 'designation' : 5, 
               'version' : 5, 
               'type'  : 5,
               'status' : 90,
               'id' : randbelow( 2  ** 64 ),
               'payload' : { 'lurk_state' : b'1234' } }

## Sending a status set to success should not be possible
## with an error message. This needs to be controlled by the server 
## as opposed to by the definition of the structure.
error_resp_unval = { 'designation' : 'lurk', 
                     'version' : 'v1', 
                     'type'  : 'invalid_',
                     'status' : 'success',
                     'id' : randbelow( 2  ** 64 ),
                     'payload' : { 'lurk_state' : b'1234' } }



for msg in [ ping_req, ping_resp, cap_req, cap_resp, error_resp, error_resp2 ]:
  test_struct( LURKMessage, msg, no_title=True, print_data_struct=True )

for msg in [ error_resp_unval ]:  
  try:
    test_struct( LURKMessage, msg, no_title=True )
  except MappingError as e :
    print( f"{e.args[ 0 ]}" )

# testing how mapping Error can be caught 
ping_req_test = deepcopy( ping_req )
ping_req_test[ 'designation' ] = 'test'
try:
  test_struct( LURKMessage, ping_req_test, no_title=True )
except MappingError as e :
  print( f"{e.args[ 0 ]}" )
  msg = e.args[ 0 ]
  m = msg.split ('\n' )
  var_type = m[ 0 ].split( '->' )
  var_type = var_type[ 1] 
  var_value = m[ 1 ].split( 'for' )
  var_value = var_value[ 1 ] 
  print( f"vartype: {var_type} - var_value: {var_value}" ) 

## testing parsing with extra bytes
## extra bytes are simply ignored during parsing
byte_ping_req = LURKMessage.build( ping_req )
LURKMessage.parse( byte_ping_req + b'\x00' )



from pylurk.cs import CryptoService
import pylurk.conf

conf = deepcopy( pylurk.conf.conf_template ) 

cs = CryptoService( conf )
for req in [ ping_req, cap_req ]:
  resp = cs.serve( LURKMessage.build( req ) )
  resp =  LURKMessage.parse( resp )
  match( req, resp ) 

byte_ping_req = LURKMessage.build( ping_req )

print( "## testing too short packet" )
byte_req = byte_ping_req[:-1]
resp = cs.serve( byte_req )
if resp != b'':
  raise ValueError( f" too short packets ( {len(byte_req)} bytes )"\
                    f"expects to receive b'' as a response." )
print( "## testing bad designation" )
byte_req = byte_ping_req
byte_req = b'\xff' + byte_ping_req[1:]
byte_resp = cs.serve( byte_req )
resp = LURKMessage.parse( byte_resp )
if resp[ 'status' ] != 'invalid_extension' :
  raise ValueError( f" 'invalid_extension' status expected {resp}" )

print( "## testing bad type" )
byte_req = byte_ping_req
byte_req = byte_ping_req[ : 2] + b'\xff' + byte_ping_req[ 3 : ]
byte_resp = cs.serve( byte_req )
resp = LURKMessage.parse( byte_resp )
if resp[ 'status' ] != 'invalid_type' :
  raise ValueError( f" 'invalid_type' status expected {resp}" )

print( "## testing bad status" )
byte_req = byte_ping_req
byte_req = byte_ping_req[ : 3] + b'\xff' + byte_ping_req[ 4 : ]
byte_resp = cs.serve( byte_req )
resp = LURKMessage.parse( byte_resp )
if resp[ 'status' ] != 'invalid_status' :
  raise ValueError( f" 'invalid_status' status expected {resp}" )


print( "## testing bad payload" )
byte_req = byte_ping_req
byte_req = byte_ping_req + b'\xff'
byte_resp = cs.serve( byte_req )
resp = LURKMessage.parse( byte_resp )
print( f"{resp}" )
if resp[ 'status' ] != 'invalid_type' :
  raise ValueError( f" 'invalid_type' status expected {resp}" )








## lurk_message handling
import pickle
#import cloudpickle
#import joblib
from cryptography.hazmat.primitives.hashes import Hash, SHA256
#from copy import deepcopy


class LURKError(Exception):
  def __init__(self, status:str,  message:str="" ):
    """ Generic Error class

    Args:
      expression : provides the python structures that generated the error
      message (str) : is a human readable message
      status (int) : the error number, that is the number returned in the LURK Header
      payload (dict) : a potential payload carried by the LURK Error Message
    """
    self.status = status
    self.message = message

class ImplementationError( LURKError ):
  def __init__(self, message:str ):
    """ Error that are related to the implementation """
    self.status = 'implementation_error'
    self.message = message

class ConfigurationError( LURKError ):
  def __init__(self, message:str ):
    """ Error that are related to the implementation """
    self.status = 'configuration_error'
    self.message = message


class LurkExt:

  def __init__( self, conf ):
    """ handles LURK messages """
    self.conf = conf
    self.lurk_state = self.get_lurk_state()

  def payload_resp( self, req: dict ) -> dict :
    if req[ 'type' ] == 'ping':
      payload = {}
    elif req[ 'type' ] == 'capabilities':
      supported_ext = []
      for ext in self.conf[ 'enabled_extensions' ]:
        supported_ext.append( { 'designation' : ext[ 0 ],
                                'version' : ext[ 1 ] } )
      payload = { 'supported_extensions' : supported_ext,
                  'lurk_state' : self.lurk_state }
    else:
      raise ImplementationError( 'invalid_type' )
    return payload

  ## NEEDS to be implemented. We are currently unable to copy,
  ## dump the self.conf object. It seems we are not able to copy
  ## the '_XXX' keys - as if we coudl not access the memory. One
  ## way may be to implement a function in the Configuration that
  ## is able to dump the self.conf...
  def get_lurk_state( self ):
    """ issue states from the configuration  """
    h = Hash( SHA256( ) )
#    conf = deepcopy( self.conf )
#    print( f" con: {conf}" )
    h.update( pickle.dumps( self.conf[ ( 'lurk', 'v1' ) ] ) )
#    h.update( cloudpickle.dumps( self.conf ) )
#    h.update( bytes( self.conf, 'utf-8' ) )
    digest = h.finalize( )
#    digest = joblib.hash( self.conf, hash_name='sha1', coerce_mmap=False)
    return digest[ : 4 ]





import secrets
import select
import socket
import selectors
import pylurk.tls13.struct_lurk_tls13
from pylurk.struct_lurk import LURKMessage, LURKHeader
from pylurk.lurk.lurk_lurk import ConfigurationError, LURKError
import pylurk.cs

class BaseTls13LurkClient :

  """ The base class for the TLS 1.3 LURK client.

  This class defines an API to format LURK requests and handle
  the LURK responses.
  In addition, this class enables the LURK Client to interacts
  with a CS that is implemented via a library.
  When used in that context, the CS MUST be provided as an object.
  The CS can be set to None when no interaction is being
  requires with the CS.
  This is mostly used in the case where classes inherite from the
  class to format the LURK requests and handle the LURK response
  but handle the communication between the LURK client and teh CS
  on their own.

  Attributes:
    conf: a dictionnary that carries all configuration parameters
    lurk_client_session_id: the session ID of the session between
      the LURK client and the LURK server (or CS). Note that the
      intent is that the client is having a single session with
      the server, and all queries are carried via this session.
      The session id consists of 4 byte randomly generated at
      the instantiation of the LURK client
    freshness: the freshness algorithm used between the LURK
      client and teh CS. BY default it is set to 'sha256'.
    cs: the instance of the CS to which the LURK client is connected.
      This parameter is only provided when the LURK client and
      the CS are having a connection type of type 'lib_cs',
      otherwise it is set to None.
  """

  def __init__( self, conf:dict={ 'connection_type' : 'lib_cs',
                                  'freshness' : 'sha256' },
                       cs=None ):

    self.conf = conf
#    self.con_type = 'lib_cs'
    self.lurk_client_session_id = secrets.token_bytes( 4 )
    self.cs_session_id  = None
    self.freshness = 'sha256'
    if 'freshness' in self.conf.keys():
      self.freshness = self.conf[ 'freshness' ]
    self.cs = cs

  def secret_req( self, **kwargs ):
    """Format secret_request field of the LURK Request

    The LURK Client expresses its request for secrets as:
    secret_request = [ 'b', 'e_s', ... ]
    In other words, only the requested secrets are mentioned.
    The LURKRequest structure required all secrets type to be
    explicitly mentioned. This is performed by this function.

    Args:
      kwargs: the key word arguments or th efunction req

    Returns:
      secret_req: the structure of the secret_request field.
    """

    secret_req = { 'b' : False, 'e_s' : False, 'e_x' : False, \
                   'h_c' : False, 'h_s' : False, 'a_c' : False, \
                   'a_s' : False, 'x' : False, 'r' : False }
#    if 'secret_request' in kwargs.keys():
    if 'secret_request' in kwargs:
#      for k in secret_req.keys():
      for k in secret_req:
        if k in kwargs[ 'secret_request' ]:
          secret_req[ k ] = True
    return secret_req

  def session_id( self, req_type, **kwargs ):
    """ handles the session id field between cs_session_id and engine_session_id
    Args:
      kwargs: the key word arguments or the function req

    Returns:
      session_id: the session_id field of the request
    """

#    if 'session_id' in kwargs.keys():
    if 'session_id' in kwargs:
      session_id = kwargs[ 'session_id' ]
    elif req_type == 'c_init_client_finished' :
      tag = self.tag( **kwargs )
      if tag[ 'last_exchange' ] is True:
        session_id = b''
      else:
        session_id =  self.lurk_client_session_id
    elif '_init_' in req_type :
      session_id =  self.lurk_client_session_id
    else:
      session_id = self.cs_session_id
    return session_id

  def fresh( self, **kwargs ):
    """handles the freshness field of the LURK request (req)

    Args:
      kwargs: the key word arguments or the function req

    Returns:
      freshness: the session_id field of the request
    """

#    if 'freshness' in kwargs.keys():
    if 'freshness' in kwargs:
      freshness = kwargs[ 'freshness' ]
      self.freshness = freshness
    else:
      freshness = self.freshness
    return freshness

  def tag( self, **kwargs ):
    """ set the tag structure ccording to the last_exchange value """
#    if 'tag' in kwargs.keys() :
    if 'tag' in kwargs :
      tag = kwargs[ 'tag' ]
#    elif 'last_exchange' in kwargs.keys():
    elif 'last_exchange' in kwargs:
      tag = { 'last_exchange' : kwargs[ 'last_exchange' ] }
    else :
      tag = { 'last_exchange' : True }
    return tag

  def req( self, req_type, **kwargs ):
    """ a common way to express a LURK Request from the LURK client

    Args:
      kwargs: the key word arguments or the function req

    Returns:
      lurk_req: a dictionnary that represenst the LURK Request
        structure
    """
    payload = {}
    if req_type == 'c_init_client_hello':
      try:
        psk_metadata_list = kwargs[ 'psk_metadata_list' ]
      except KeyError:
        psk_metadata_list = [] ## no psk
      try:
        secret_request = kwargs[ 'secret_request' ]
      except KeyError:
        kwargs[ 'secret_request' ] = [ 'b', 'e_s', 'e_x' ]

      payload = { \
        'session_id' : self.session_id( req_type, **kwargs ),
        'handshake' : kwargs[ 'handshake' ],
        'freshness' : self.fresh( **kwargs ) ,
        'psk_metadata_list' : psk_metadata_list,
        'secret_request' : self.secret_req( **kwargs )
      }
    elif req_type == 'c_server_hello':
      payload = { \
        'session_id' : self.session_id( req_type, **kwargs ),
        'handshake' : kwargs[ 'handshake' ],
        'ephemeral' : kwargs[ 'ephemeral' ]
      }
    elif req_type == 'c_client_finished':
      payload = { \
        'tag' : self.tag( **kwargs ),
        'session_id' : self.session_id( req_type, **kwargs ),
        'handshake' : kwargs[ 'handshake' ],
        'server_certificate' : kwargs[ 'server_certificate' ],
        'client_certificate' : kwargs[ 'client_certificate' ],
        'secret_request' : self.secret_req( **kwargs )
      }
    elif req_type == 'c_register_tickets':
      payload = { \
        'tag' : self.tag( **kwargs ),
        'session_id' : self.session_id( req_type, **kwargs ),
        'ticket_list' : kwargs[ 'ticket_list' ]
      }
    elif req_type == 'c_init_client_finished' :
      payload = { \
        'tag' : self.tag( **kwargs ),
        'session_id' : self.session_id( req_type, **kwargs ),
        'handshake' : kwargs[ 'handshake' ],
        'server_certificate' : kwargs[ 'server_certificate' ],
        'client_certificate' : kwargs[ 'client_certificate' ],
        'freshness' : self.fresh( **kwargs ) ,
        'ephemeral' : kwargs[ 'ephemeral' ],
        'psk' : kwargs[ 'psk' ]
      }
    elif req_type in [ 'ping', 'capabilities' ]:
      payload = {}
    else:
      raise ValueError( f"Unknown request type {req_type}" )
    lurk_req = \
      { 'designation' : 'tls13',
        'version' : 'v1',
        'type' : req_type,
        'status' : 'request',
        'id' : secrets.randbelow( 2  ** 64 ), ## MUST be int
        'payload' : payload }
    if req_type in [ 'ping', 'capabilities' ]:
      lurk_req [ 'designation' ] = 'lurk'
    return lurk_req

  def bytes_resp( self, bytes_req:bytes ):
    """ return the byte response with the appropriated transport

    Args:
      - bytes_req: the lurk response expressed in bytes.
    """
    return self.cs.serve( bytes_req )


  def resp( self, req_type, **kwargs ):
    lurk_req = self.req( req_type, **kwargs )
#    print( f"____ {lurk_req}" )
    print( f"--- E -> CS: Sending {req_type} Request:" )
#    print( f"--- E -> CS: {lurk_req }" )
#    if lurk_req[ 'type' ] == 'c_server_hello' :
#      print( f"--- E -> CS: ephemral { pylurk.tls13.struct_lurk_tls13.Ephemeral.build( lurk_req[ 'payload' ][ 'ephemeral' ], _status='request', )}" )
#      print( f"  - build: {LURKMessage.build( lurk_req )}" )
#    print( f"  - {LURKMessage.parse( LURKMessage.build( lurk_req ) )}" )
#    print( f" lurk_req: {lurk_req}" )
#    print( f" build: {LURKMessage.build( lurk_req )}" )
#    print( f" bytes_resp : {self.bytes_resp( LURKMessage.build( lurk_req ) )}" )
    lurk_resp = LURKMessage.parse( self.bytes_resp( LURKMessage.build( lurk_req ) ) )
#    print( f"resp: {lurk_resp}" )
    if lurk_resp[ 'status' ] != 'success':
      raise LURKError( lurk_resp[ 'status' ],  f"Lurk exchange error: {lurk_resp}" )
    print( f"--- E <- CS: Receiving {req_type} Response:" )
#    print( f"  - {LURKMessage.parse( LURKMessage.build( lurk_resp ) )}" )
    ## updating the session_id (sending)
    if '_init_' in req_type :
      try:
        self.cs_session_id = lurk_resp[ 'payload' ][ 'session_id' ]
      except:
        pass
    return lurk_resp

#    if lurk_resp[ 'status' ] != 'success':
#      raise ValueError( f"Lurk exchange error: {lurk_resp}" )
#    print( "--- E <- CS: Receiving {req_type} Response:" )
#    print( f"  - {LURKMessage.parse( LURKMessage.build( lurk_resp ) )}" )
#    ## updating the session_id (sending)
#    if '_init_' in req_type :
#      try:
#        self.cs_session_id = lurk_resp[ 'payload' ][ 'session_id' ]
#      except:
#        pass
#    return lurk_resp

class TCPTls13LurkClient( BaseTls13LurkClient ) :

  def __init__( self, conf:dict ):
    """ configures the LURK client in a stateless TCP mode

    Stateless TCP means that every message is sent via a
    specific newly established TCP session

    - conf designates the configuration parameters for the CS.
      The complete configuration of the CS can be used, but only
      a subset of the parameters are being used.
      Mostly fqdn - ip_address and port.

    """
    super().__init__( conf=conf, cs=None )
#    self.conf = conf
#    self.conf_type = 'stateless_tcp'
    self.server_address = self.get_server_address_from_conf( )

  def get_server_address_from_conf( self ):
    """ return host and port from the configuration file """
#    print( self.conf.keys() )
#    key_list = self.conf[ 'connectivity' ].keys()
#    if 'type' in key_list:
#      cs_type = self.conf[ 'connectivity' ][ 'type' ]
#      if cs_type != 'stateless_tcp':
#        raise ConfigurationError( f"unexpected type {cs_type} for "\
#          f"{self.__class__.__name__}. Expecting 'stateless_tcp'." )
#
#else:
#      raise ConfigurationError( f"Cannot find type in configuration "\
#        f"{self.__class__.__name__}. Expecting 'stateless_tcp'." )
    host = None
    con_conf = self.conf[ 'connectivity' ]
    if 'fqdn' in con_conf.keys():
      fqdn = con_conf[ 'fqdn' ]
      if fqdn not in [ None, '' ]:
        ## maybe we coudl perform a DNS lookup
        host = fqdn
    if host is None and 'ip' in con_conf.keys():
      host = con_conf[ 'ip' ]
    else:
      raise ConfigurationError( f"Cannot find 'ip' or 'fqdn' in "\
        f"configuration: {self.conf} for {self.__class__.__name__}." )
    if 'port' in con_conf.keys():
      port = con_conf[ 'port' ]
    else:
      raise ConfigurationError( f"Cannot find 'port' in "\
        "configuration: {self.conf} for {self.__class__.__name__}." )
    return ( host, port )

#  def resp( self, req_type, **kwargs ):
#
#    lurk_req = self.req( req_type, **kwargs )
#    print( f"--- E -> CS: Sending {req_type} Request:" )
#    lurk_req_bytes = LURKMessage.build( lurk_req )
#    lurk_resp = LURKMessage.parse( self.cs.serve( LURKMessage.build( lurk_req ) ) )
  def bytes_resp( self, bytes_req:bytes ):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
      # Connect to server and send data
      sock.connect( self.server_address )
      sock.sendall( bytes_req )
      return sock.recv(1024)


class PersistentTCPTls13LurkClient( TCPTls13LurkClient ) :

  def __init__( self, conf:dict ):
    """ configures the LURK client in a persitent TCP mode

    While the TCP session remains persistent and thus can
    be re-used for multiple exchanges, this TCP client
    does not implement ways to pipeline query and responses.
    To do so would require a sort of Lurk resolver / proxy
    that binds response received to queries.
    """
    super().__init__( conf=conf )
    self.sock = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
    self.sock.connect( self.server_address )
    self.selector = selectors.DefaultSelector()
    ## registering the "accept" socket
    self.selector.register(fileobj=self.sock, \
                           events=selectors.EVENT_READ )

#    self.serve_forever( )

  def bytes_resp( self, bytes_req:bytes ):
    ## this should work
    self.sock.sendall( bytes_req )
    return self.sock.recv(1024)


    ## alternatively we may parse the exact response.
    ## introducing some listening
    ## we need thus to define two threads: one listening
    ## to the cs the other to stub clients.
  def server_forever( self ):
    bytes_recv = b''
    while len(bytes_recv) < pylurk.cs.HEADER_SIZE + 4 :
      rlist, wlist, xlist = select.select([self.sock], [], [])
      if len(rlist) > 0:
        bytes_recv = self.sock.recv( pylurk.cs.HEADER_SIZE + 4 )
    header = LURKHeader.parse( bytes_recv )
    bytes_nbr = header['length']
    print( f"DEBUG: CS: bytes_nbr: {bytes_nbr}" )
    bytes_recv += self.sock.recv(min(bytes_nbr - len(bytes_recv), 4096))
    return bytes_recv


def get_lurk_client_instance( conf:dict=None, cs=None ):
  """ returns a LURK Client instance as defined by the configuration """

  if conf is None:
    conf = { 'connectivity' : { 'type' : 'lib_cs' } }
  con_type = conf[ 'connectivity' ] [ 'type' ]
  if con_type == 'lib_cs' :
    if cs is None:
      raise ValueError( "cs MUST be provided " )
    lurk_client =  BaseTls13LurkClient( conf, cs=cs )
  elif con_type == 'tcp':
    lurk_client = TCPTls13LurkClient( conf )
  elif con_type == 'persistent_tcp':
    lurk_client = PersistentTCPTls13LurkClient( conf )
  else:
    raise ConfigurationError( f"unknown connection_type {con_type}" )
  return lurk_client



import sys
import traceback
from copy import deepcopy
import json

from secrets import token_bytes
from cryptography.hazmat.primitives.hashes import Hash, SHA256

## inserting the src in the search path
import sys
sys.path.insert( 0, '../src/')

from pylurk.tls13.struct_tls13 import *
from test_utils import *

## signature_algorithms
sig_list = [
  'rsa_pkcs1_sha256', 
  'rsa_pkcs1_sha384',
  'ecdsa_secp256r1_sha256', 
  'ecdsa_secp384r1_sha384',
  'ed25519', 'ed448'
  ]

sig_scheme_list = { 'supported_signature_algorithms' : sig_list}
test_struct(SignatureSchemeList, sig_scheme_list)

ext13 = {'extension_type': 'signature_algorithms', \
         'extension_data' : sig_scheme_list }
test_struct(Extension, ext13)

## psk_key_exchange_modes
psk_modes = {'ke_modes' : ['psk_ke', 'psk_dhe_ke']}
test_struct( PskKeyExchangeModes, psk_modes)

ext45 = {'extension_type': 'psk_key_exchange_modes', \
         'extension_data' : psk_modes }
test_struct(Extension, ext45)

## pre_shared_key
psk_id = {'identity' : b'\x00\x00', \
          'obfuscated_ticket_age' : b'\x00\x01\x02\x03' }
psk_binder = {'binder' : b'\xff\xff\xff\xff'}
offered_psks= { 'identities' : [psk_id, psk_id], \
                'binders' : [psk_binder, psk_binder]}
test_struct(OfferedPsks, offered_psks)

ext41_ch = { 'extension_type': 'pre_shared_key', \
             'extension_data' : offered_psks }
test_struct(Extension, ext41_ch, ctx_struct={'_msg_type' : 'client_hello' } )

ext41_sh =  { 'extension_type': 'pre_shared_key', \
              'extension_data' : 0 }
test_struct(Extension, ext41_sh, ctx_struct={'_msg_type' : 'server_hello' } )

## post-handshake authentication
ext49 = {'extension_type': 'post_handshake_auth', \
         'extension_data' : {} }
test_struct(Extension, ext49)

## supported_group

grp = {'named_group_list' : ['secp256r1', 'secp384r1', 'x25519', 'x448' ]}

test_struct(NamedGroupList, grp)

ext10 = {'extension_type': 'supported_groups', \
         'extension_data' : grp }
test_struct(Extension, ext10)

## key_share
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend

private_key = ec.generate_private_key( ec.SECP256R1(), default_backend())
public_key = private_key.public_key()
public_numbers = public_key.public_numbers()

x = public_numbers.x
y = public_numbers.y

print("Public Numbers: %s"%public_numbers)
print("  - x: %s"%x)
print("  - x: %s"%(x).to_bytes(32, byteorder='big'))
print("  - y: %s"%y)
print("  - y: %s"%(y).to_bytes(32, byteorder='big'))

secp256r1_key = { 'legacy_form' : 4, 'x' : x, 'y' : y }
ke_entry_secp256r1 = {'group': 'secp256r1', 'key_exchange' : secp256r1_key}

private_key = ec.generate_private_key( ec.SECP384R1(), default_backend())
public_key = private_key.public_key()
public_numbers = public_key.public_numbers()

x = public_numbers.x
y = public_numbers.y

print("Public Numbers: %s"%public_numbers)
print("  - x: %s"%x)
print("  - x: %s"%(x).to_bytes(48, byteorder='big'))
print("  - y: %s"%y)
print("  - y: %s"%(y).to_bytes(48, byteorder='big'))

secp384r1_key = { 'legacy_form' : 4, 'x' : x, 'y' : y }
ke_entry_secp384r1 = {'group': 'secp384r1', 'key_exchange' : secp384r1_key}

private_key = ec.generate_private_key( ec.SECP521R1(), default_backend())
public_key = private_key.public_key()
public_numbers = public_key.public_numbers()

x = public_numbers.x
y = public_numbers.y

print("Public Numbers: %s"%public_numbers)
print("  - x: %s"%x)
print("  - x: %s"%(x).to_bytes(66, byteorder='big'))
print("  - y: %s"%y)
print("  - y: %s"%(y).to_bytes(66, byteorder='big'))
secp521r1_key = { 'legacy_form' : 4, 'x' : x, 'y' : y }
ke_entry_secp512r1 = {'group': 'secp521r1', 'key_exchange' : secp521r1_key}


from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives import serialization

private_key = X25519PrivateKey.generate()
public_key = private_key.public_key()
x25519_key = public_key.public_bytes(
          encoding=serialization.Encoding.Raw,
          format=serialization.PublicFormat.Raw)
ke_entry_x25519 = {'group': 'x25519', 'key_exchange' : x25519_key}

from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey

private_key = X448PrivateKey.generate()
public_key = private_key.public_key()
x448_key = public_key.public_bytes(
          encoding=serialization.Encoding.Raw,
          format=serialization.PublicFormat.Raw)
ke_entry_x448 = {'group': 'x448', 'key_exchange' : x448_key}

ke_entries = [ke_entry_secp256r1,
              ke_entry_secp384r1, 
              ke_entry_secp512r1, 
              ke_entry_x25519,
              ke_entry_x448 ]

ks_ch = {'client_shares' : ke_entries}
test_struct(KeyShareClientHello, ks_ch)

ks_hr = {'selected_group' : 'x448' }
test_struct(KeyShareHelloRetryRequest, ks_hr)

for ke_entry in ke_entries:
  ks_sh = {'server_share' : ke_entry }
  test_struct(KeyShareServerHello, ks_sh)

empty_ke_entry = {'group': 'x448', 'key_exchange': b'' }
test_struct( EmptyKeyShareEntry, empty_ke_entry )  

ks_sh_empty = { 'server_share': empty_ke_entry }
test_struct( KeyShareServerHelloEmpty, ks_sh_empty )


ext51_ch = {'extension_type': 'key_share', \
            'extension_data' : ks_ch }
ctx_struct = {'_msg_type' : 'client_hello'}
test_struct(Extension, ext51_ch, ctx_struct=ctx_struct)

ext51_hr = {'extension_type': 'key_share', \
            'extension_data' : ks_hr }
ctx_struct = {'_msg_type' : 'server_hello'}
test_struct(Extension, ext51_hr, ctx_struct=ctx_struct)

ext51_sh = {'extension_type': 'key_share', \
            'extension_data' : ks_sh }
ctx_struct = {'_msg_type' : 'server_hello'}
test_struct(Extension, ext51_sh, ctx_struct=ctx_struct)

ext51_sh_empty = {'extension_type': 'key_share', \
                  'extension_data' : ks_sh_empty }
ctx_struct = {'_msg_type' : 'server_hello'}
test_struct(Extension, ext51_sh_empty, ctx_struct=ctx_struct)


## III TLS messages
## Some Extensions have different structures depending on the msg_type
## parameter. As a result, it is preferred to test the message via the
##  Handshake structure. 

## client_hello
random = token_bytes( 32 )
client_hello = {\
  'legacy_version' : b'\x03\x03',
  'random' : random,
  'cipher_suites' : ['TLS_AES_128_GCM_SHA256', 'TLS_CHACHA20_POLY1305_SHA256'],
  'legacy_compression_methods' : b'\x00',
  'extensions' : [ext49]
  }
test_struct(ClientHello, client_hello)

client_hello['extensions'].append(ext51_ch)
ctx_struct = {'msg_type': 'client_hello'}
#test_struct(ClientHello, client_hello, ctx_struct=ctx_struct)

hs_client_hello = {'msg_type': 'client_hello', 'data' : client_hello}
test_struct(Handshake, hs_client_hello)

## server_hello

server_hello = {
  'legacy_version' : b'\x03\x03',
  'random' : random,
  'cipher_suite' :'TLS_AES_128_GCM_SHA256',
  'legacy_compression_method' : b'\x00',
  'extensions' : [ext49, ext51_sh]
  }
hs_server_hello = {'msg_type': 'server_hello', 'data' : server_hello}
## unexplained error with check
test_struct(Handshake, hs_server_hello)

# end_of_early_data
hs_end_of_early_data = {
  'msg_type' : 'end_of_early_data', 
  'data' : {}}
test_struct(Handshake, hs_end_of_early_data)

##encrypted_extensions
hs_encrypted_extensions = {
  'msg_type' : 'encrypted_extensions', 
  'data' : { 'extensions' :  [ ext49, ext13, ext45 ] }
}
test_struct(Handshake, hs_encrypted_extensions)

## certificate_request

hs_certificate_request = {
  'msg_type' : 'certificate_request', 
  'data' : { 'certificate_request_context' :  b'\x00\x01',
             'extensions' : [ ext49, ext13, ext45 ] }

}
test_struct(Handshake, hs_certificate_request)


## certificate
cert_entry = {'cert' : b'\x00\x01\x02\x03',\
              'extensions':[] }
cert = { 'certificate_request_context' : b'\x00\x01', 
         'certificate_list' : [ cert_entry, cert_entry, cert_entry ] }
ctx_struct = { '_certificate_type' : 'X509' }
test_struct(Certificate, cert, ctx_struct=ctx_struct)

hs_certificate = {
  'msg_type' : 'certificate', 
  'data' : { 'certificate_request_context' : b'\x00\x01', 
             'certificate_list' : [ cert_entry, cert_entry, cert_entry ] }
}
ctx_struct = { '_certificate_type' : 'X509' }
test_struct(Handshake, hs_certificate, ctx_struct=ctx_struct)


## certificate_verify
hs_certificate_verify = {
  'msg_type' : 'certificate_verify',
  'data' : { 'algorithm' : 'ed25519', 
             'signature' : b'\x00\x01\x02' }
}
test_struct(Handshake, hs_certificate_verify)

## finished
hs_finished = {
  'msg_type' : 'finished', 
  'data' : {'verify_data' : token_bytes( 32 )}
}
ctx_struct = { '_cipher' : 'TLS_AES_128_GCM_SHA256' }
test_struct(Handshake, hs_finished, ctx_struct=ctx_struct)

## NewSessionTicket
new_session_ticket = { \
  'ticket_lifetime':5,\
  'ticket_age_add':6,\
  'ticket_nonce':b'\x07', \
  'ticket':b'\x00\x01\x02\x03',\
  'extensions':[]\
}

hs_new_session_ticket = {'msg_type' : 'new_session_ticket', 
                         'data' : new_session_ticket }
test_struct(Handshake, hs_new_session_ticket)


## key_update
hs_key_update = {
  'msg_type' : 'key_update', 
  'data' : {'request_update' : 'update_requested' }
}
test_struct(Handshake, hs_key_update)
  



import os
import os.path
from  os.path import join
import pathlib
import shutil
import datetime
import argparse
#import pkg_resources
#import copy
#from copy import deepcopy
#import hashlib ## mostly for compatibility with key_schedule

from cryptography.hazmat.backends import default_backend
#from cryptography.hazmat.primitives.asymmetric.x25519 import \
#  X25519PrivateKey, X25519PublicKey
#from cryptography.hazmat.primitives.asymmetric.x448 import \
#  X448PrivateKey, X448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import \
  Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import \
  Ed448PrivateKey, Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import \
  RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import \
  EllipticCurvePrivateKey, EllipticCurvePublicKey
#  SECP256R1, SECP384R1, SECP521R1, ECDSA, \
#  EllipticCurvePublicNumbers, ECDH,
from cryptography.hazmat.primitives.hashes import \
  Hash, SHA256
# SHA384, SHA512
from cryptography.hazmat.primitives.serialization import \
  load_der_private_key, load_der_public_key, load_der_public_key, \
  NoEncryption, Encoding, PrivateFormat, PublicFormat
# load_pem_private_key, \
from cryptography import x509
from cryptography.x509.oid import NameOID
#from cryptography.hazmat.primitives.asymmetric import padding
#from cryptography.hazmat.primitives.ciphers import \
#  Cipher, algorithms, modes
#from cryptography.hazmat.primitives.ciphers.aead import \
#  ChaCha20Poly1305, AESGCM, AESCCM
#from cryptography.hazmat.primitives.kdf.hkdf import \
#  HKDF, HKDFExpand

from construct.core import MappingError

import pytls13.struct_tls13 as tls

from pylurk.lurk.lurk_lurk import ConfigurationError
#  ImplementationError
# LURKError, \
#import pylurk.tls13.struct_tls13 as lurk
import pylurk.tls13.crypto_suites
#import binascii



class Configuration:
  """manipulates configuration file

    configuration class has two distinct goals:
      1. easing the use interactions with the configuration
        dictionary ( fill, remove items ). As to avoid
        misconfiguration, the initial configuration file expect
        to carry the minimal amount of information. For example,
        public key information is filled at one place, and any
        arguments are them derived from it - even tough this
        might require extra processing.

      2. generating informations that can be used by the
        software: This typically includes some information that
        can be derive. In this case, we would like to avoid
        information to be reprocessed muliutple times. For
        example, we do not want the certificate to read from the
        file each time a CERT message is sent.
        For that purpose we are adding some extra variables that
        are noted _key to distinguish from those expected from
        the user.

    Attributes:
      conf: a dictionary with all configuration parameters
  """

  def __init__( self ):
    """ initializes self.conf """

#    data_dir = pkg_resources.resource_filename(__name__, '../data/')
    self.conf = {
      'profile' : 'explicit configuration',
      'description' : "LURK Cryptographic Service configuration template",
      ## connectivity describes how to connect the CS
      ##  type: describe sthe type of connectivity.
      ##    Values include 'lib_cs', stateless_tcp, tcp, tcp+tls, http, https,..
      ##    The default type is lib_cs. other types are usually associated with
      ##    other parameters that needs to be provided explicitly
      'connectivity' : {
          'type' : "lib_cs"
           #'fqdn' : None,
           #'ip_address' : "127.0.0.1",
           #'port' : 9999,
           #'key' : join(data_dir, 'key_tls12_rsa_server.key'),
           #'cert' : join(data_dir, 'cert_tls12_rsa_server.crt'),
           #'key_peer' : join(data_dir, 'key_tls12_rsa_client.key'),
           #'cert_peer' : join(data_dir, 'cert_tls12_rsa_client.crt')
          },
      'enabled_extensions' : [ ( 'lurk', 'v1' ) , ( 'tls13', 'v1' ) ],
      ( 'lurk', 'v1' ) : {
         'type_authorized' : [ 'ping',  'capabilities' ]
      },
      ( 'tls13', 'v1' ) : {
        ## prints multiple useful information
        ## test_vector is not part of the default template and
        ## is expected to be explicitly provided.
        ##   trace (bool) : indicate to provide multiple information
        ##   file : is te path tho the file that is use to record the
        ##     parameters or to check the measured parameters against those
        ##     provided in the test vector
        'debug' : {
          'trace' : True,
          #'test_vector' : {
          #'file' : './illustrated_tls13.json',
          #'mode' : 'check', # possible value None /check / record
          #   }
        },
        ## role indicates if the CS is implemented for a
        ## TLS client or a TLS server
        'role' : 'server', #[ 'client', 'server' ],
        ## defines the type of LURK requests that are served by the
        ## CS.
        'type_authorized' : [ 's_init_cert_verify',     ## server ECDHE
                              's_init_early_secret',    ## server PSK
                              's_hand_and_app_secret',  ## server PSK
                              's_new_ticket',           ## session resumption
                              'c_init_client_finished', ## only signature
                              'c_post_hand_auth',       ## post hand
                              'c_init_client_hello',    ## client hello
                              'c_server_hello',
                              'c_client_finished',
                              'c_register_tickets' ],
        ## echde authentication parameter
        ## certificate chain, the public key of the TLS client or server
        ## is the last one.
        ##
        ## Note that the file name follows a generic patent to specify
        ## the type of the key (in some case sig_scheme) as well as the
        ## type of encoding.
        ## This information are not necessary, and we shoudl be able to
        ## remove that constraint.
        ##
        ## Note that the chain of certificates is likely to be stored
        ## into a single file.
        #'public_key' : [join( data_dir, 'cert-rsa-enc.der' )],
        #'private_key' : join( data_dir, 'key-rsa-enc.pkcs8' ), ## der, pkcs8
         'ephemeral_method_list' : ['no_secret', 'cs_generated', 'e_generated'],
         'authorized_ecdhe_group' : ['secp256r1', 'secp384r1',
                                     'secp521r1', 'x25519', 'x448'],
         ## only one signature scheme must be selected
         'sig_scheme' : ['rsa_pkcs1_sha256', \
                       'rsa_pkcs1_sha384', \
                       'rsa_pkcs1_sha512',\
                       'ecdsa_secp256r1_sha256', \
                       'ecdsa_secp384r1_sha384',\
                       'ecdsa_secp521r1_sha512', \
                       'rsa_pss_rsae_sha256', \
                       'rsa_pss_rsae_sha384', \
                       'rsa_pss_rsae_sha512', \
                       'ed25519', \
                       'ed448', \
                       'rsa_pss_pss_sha256', \
                       'rsa_pss_pss_sha384', \
                       'rsa_pss_pss_sha512' ],
         ## secrets
         'client_early_secret_authorized' : True,    ## only for PSK
         'early_exporter_secret_authorized' : True,
         'exporter_secret_authorized' : True,
         'app_secret_authorized' : True,             ## anytime application is protected by TLS
         'resumption_secret_authorized' : True,     ## PSK / non protected ECDHE

         ## machinery logic
         's_init_early_secret_session_id' : True,
         'last_exchange' : { 's_init_cert_verify' : False,
                             's_hand_and_app_secret' : False,
                             'c_init_client_finished' : False,
                             'c_init_post_auth' : False,
                             'c_client_finished' : False},
         ## ticket related configuration
         'max_tickets' : 6,
         'ticket_life_time' : 172800, # 2d = 2*24*3600 < 2**32-1
         'ticket_nonce_len' : 20,  ## bytes < 255
         'ticket_generation_method': 'ticket', ## versus index
         ## the key used to encrypt tickets
         ##
         #'ticket_public_key': join(data_dir, 'ticket-cert-rsa-enc.der'), ## one key
         #'ticket_private_key' : join( data_dir, 'key-rsa-enc.pkcs8'), ## der, pkcs8
         'ticket_len' : 4,  ## bytes < 255
         ## client parameters
         'post_handshake_authentication' : True,
         'max_post_handshake_authentication' : 1, ## 'c_post_hand_auth' ?
            },
    }

  def set_role( self, role:str ):
    """ set the role in the configuration

    Args:
      role: set the configuration to TLS client or TLS server. 
        rold must be either 'client' or 'server'
    """

    if isinstance( role, str ):
      if role not in [ 'client', 'server' ]:
        raise ConfigurationError( f"unknown role {role}." )
      # role = role
    elif isinstance( role, list ):
      for r in role:
        if r not in [ 'client', 'server' ]:
          raise ConfigurationError( f"unknown role {role}." )
    else:
      raise ConfigurationError( f"unknown role {role}. Must be string or list." )
    self.conf[ ( 'tls13', 'v1' ) ] [ 'role' ] = role

  def merge( self, branch:dict, master:dict=None ):
    """ merge th ebranch conf to the master conf

    The branch configuration is expected to provide a subset of the parameters
    Those not provided are taken from the template.

    Args:
      branch: the dictionary containing the configuration 
        parameters to be integrated
      master: the dictionary to which the branch is merged.
    """

    if master is None :
      master = self.conf
#    branch = copy.deepcopy( branch )
    for key in branch.keys():
      value_is_dict = isinstance( branch[ key ], dict )
      if value_is_dict is False or key not in master.keys() :
        master[ key ] = branch[ key ]
      else:
        master[ key ] = self.merge( branch[ key ], master[ key] )
    return master

  def generate_keys( self, tls_sig_scheme:str ):
    """ updates conf to serve the TLS signature scheme

    Args:
      tls_sig_scheme: the TLS signature scheme
    """

    try:
      tls.SignatureScheme.build( tls_sig_scheme )
    except MappingError as exc:
      raise ConfigurationError( f"Invalid sig_scheme - only TLS1.3 sig_scheme" \
            f"are expected: {self.conf[ ( 'tls13', 'v1' ) ][ 'sig_scheme' ]}." )

    sig_scheme = pylurk.tls13.crypto_suites.SigScheme( tls_sig_scheme )
    sig_algo = sig_scheme.algo
    if sig_algo == 'ed25519' :
      private_key = Ed25519PrivateKey.generate()
    elif sig_algo == 'ed448' :
      private_key = Ed448PrivateKey.generate()
    elif sig_algo == 'ecdsa' :
      private_key = ec.generate_private_key(
        sig_scheme.curve, default_backend())
    elif 'rsa' in sig_algo :
      private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048,
        backend=default_backend())
    else:
      raise ConfigurationError( f"unable to generate keys for {sig_algo.algo}" )
    public_key = private_key.public_key( )
    return private_key, public_key

  def store_keys( self, key, key_format, conf_dir ) -> list: # type 'X509', 'Raw'
    """ store key(s) and return file names

    Args:
      key: the key object. When a private key is provided private
        and public keys are stored. ( RSAPrivateKey,
        RSAPublicKey, Ed25519PrivateKey, d25519PublicKey,
        Ed448PrivateKey, Ed448PublicKey, EllipticCurvePrivateKey,
        EllipticCurvePublicKey).
      key_format: the format of the storage ( Raw or 'X509' )

    Returns:
      file_name list
    """
# changed form pylint
#    if isinstance( key, RSAPrivateKey ) or \
#       isinstance( key, Ed25519PrivateKey ) or \
#       isinstance( key, Ed448PrivateKey ) or \
#       isinstance( key, EllipticCurvePrivateKey ) :
    if isinstance( key, ( RSAPrivateKey, Ed25519PrivateKey, \
                        Ed448PrivateKey, EllipticCurvePrivateKey ) ) :
      private = True
# changed form pylint
#    elif isinstance( key, RSAPublicKey ) or\
#         isinstance( key, Ed25519PublicKey ) or \
#         isinstance( key, Ed448PublicKey ) or \
#         isinstance( key, EllipticCurvePublicKey ) :
    elif isinstance( key, ( RSAPublicKey,  Ed25519PublicKey, \
                          Ed448PublicKey, EllipticCurvePublicKey ) ) :
      private = False
    else:
      raise ConfigurationError( f"unknown key class {key.__class__.__name__}" )
    algo = self.key_algo( key )

    private_file = None
    public_file = None
    if private is True:
      class_name = key.__class__.__name__
      private_file = join( conf_dir, f"{class_name}-{algo}-pkcs8.der" )
      ## storing private key
      with open( private_file, 'wb' ) as f:
        f.write( key.private_bytes( encoding=Encoding.DER, \
          format=PrivateFormat.PKCS8, encryption_algorithm=NoEncryption() ) )
      ## retrieving the public key
      private_key = key
      public_key = private_key.public_key( )
    else:
      public_key = key
    ## storing public key
    ## Note that storing self signed certificate requires the private key
    class_name = public_key.__class__.__name__
    public_file = join( conf_dir, f"{class_name}-{algo}-{key_format}.der" )
    if key_format == 'Raw':
      with open( key, 'wb' ) as f:
        f.write( public_key.public_bytes( encoding=Encoding.DER,  \
                 format=PublicFormat.SubjectPublicKeyInfo))
    elif key_format == 'X509':
      one_day = datetime.timedelta(1, 0, 0)
      today = datetime.datetime.today()
      builder = x509.CertificateBuilder()
      builder = builder.subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, u'cryptography.io'),\
        ]))
      builder = builder.issuer_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, u'cryptography.io'),\
        ]))
      builder = builder.not_valid_before( today - one_day )
      builder = builder.not_valid_after( today + ( one_day * 30 ) )
      builder = builder.serial_number(x509.random_serial_number())
      builder = builder.public_key( public_key )
      builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(u'cryptography.io')] ),\
          critical=False )
      builder = builder.add_extension( x509.KeyUsage( \
                  digital_signature=True,
                  content_commitment=True,
                  key_encipherment=False,
                  data_encipherment=True,
                  key_agreement=False,
                  key_cert_sign=True,
                  crl_sign=False,
                  encipher_only=False,
                  decipher_only=False ), critical=False )
      ### CA
      ## Currently set to self-signed
      if private is True:
#        if isinstance( private_key, Ed25519PrivateKey ) or\
#             isinstance( private_key, Ed448PrivateKey ):
        if isinstance( private_key, ( Ed25519PrivateKey, Ed448PrivateKey ) ):
          ca_algo = None
        else:
          ca_algo = SHA256()
        builder = builder.add_extension(
          x509.BasicConstraints(ca=False, path_length=None), critical=True, )
        certificate = builder.sign (private_key=private_key, algorithm=ca_algo,\
          backend=default_backend() )
        with open( public_file, 'wb' ) as f:
          f.write( certificate.public_bytes( Encoding.DER ) )
      else:
        raise ConfigurationError( "Cannot store public key. self signed " +\
                "certificate requires the private key to be provided" )
    else:
      raise ConfigurationError( f"invalid key format {key_format}")
    return private_file, public_file

  def load_public_key( self, public_file) :
    """ load the public key and define the key format"""

    pylurk.debug.check_file( public_file )
    with open( public_file, 'rb' )  as f:
      public_bytes = f.read()
    ## trying to load the certificate
    try:
      cert = x509.load_der_x509_certificate( public_bytes, default_backend())
      public_key = cert.public_key()
      cert_type = 'X509'
    except:
      try:
        cert = x509.load_pem_x509_certificate( public_bytes, default_backend())
        public_key = cert.public_key()
        cert_type = 'X509'
      except:
        ## trying to load the raw public key
        try:
          public_key = Ed25519PublicKey.from_public_bytes( public_bytes )
          cert_type = 'Raw'
        except:
          try:
            public_key = Ed448PublicKey.from_public_bytes( public_bytes )
          except:
            ## RSAPublicKey, EllipticCurvePublicKey,
            try:
              public_key = load_pem_public_key( public_bytes, backend=default_backend() )
              cert_type = 'Raw'
            except:
              try:
                public_key = load_der_public_key( public_bytes, backend=default_backend() )
                cert_type = 'Raw'
              except:
                raise ConfigurationError( \
                  f"Unable to load public key public_file {public_file}" )
    return public_key, cert_type

  def load_private_key( self, private_file ):
    """ returns the private key """

    pylurk.debug.check_file( private_file )
    with open( private_file, 'rb' )  as f:
      private_bytes = f.read()
      ## ed25519
      try:
        private_key = Ed25519PrivateKey.from_private_bytes( private_bytes )
      except:
      ## ed448
        try:
          private_key  = Ed448PrivateKey.from_private_bytes( private_bytes )
        except:
        ## RSA / ECDSA
          try:
            private_key = load_der_private_key(private_bytes, \
                    password=None, backend=default_backend() )
          except:
            raise ConfigurationError( \
              f"Unable to load the private key from {private_file}" )
    return private_key

  def key_algo( self, key ) -> str :
    """ returns the key signature algorithm
    """
#    if isinstance( key, Ed25519PrivateKey ) or\
#       isinstance( key, Ed25519PublicKey ) :
    if isinstance( key, ( Ed25519PrivateKey, Ed25519PublicKey ) ) :
      algo = 'ed25519'
#    elif isinstance( key, Ed448PrivateKey ) or\
#        isinstance( key, Ed448PublicKey ) :
    elif isinstance( key, ( Ed448PrivateKey, Ed448PublicKey ) ) :
      algo  = 'ed448'
#    elif isinstance( key, EllipticCurvePrivateKey) or\
#        isinstance( key, EllipticCurvePublicKey ) :
    elif isinstance( key, ( EllipticCurvePrivateKey, EllipticCurvePublicKey ) ) :
      algo = 'ecdsa_' + key.curve.name
#    elif isinstance( key, RSAPrivateKey ) or\
#        isinstance( key, RSAPublicKey ) :
    elif isinstance( key, ( RSAPrivateKey, RSAPublicKey ) ) :
      algo = 'rsa'
    else:
      raise ConfigurationError( f"unknown algo for {key.__class__.__name__}" )
    return algo


  ### TLS or LURK engine related functions

  def load_cert_entry_list( self ):
    """ returns the certificate list """
    cert_file_list = self.conf[ ( 'tls13', 'v1' )  ] ['public_key']
    cert_entry_list = []
    for cert_file in cert_file_list:
      with open( cert_file, 'rb' ) as f:
        public_bytes = f.read()
      cert_entry_list.append( { 'cert' : public_bytes, 'extensions': [] } ) ## certificateEntry
    return cert_entry_list

  def set_tls13_keys( self, private_key_file=None, public_key_file=None, \
                      sig_scheme:str='ed25519', key_format='X509' ):
    """ configures the cryptographic material for the TLS 1.3 extension

    This function generates the appropriated cryptographic material when
    needed.
    This includes when the material is not specified or when not coherent.

    Args:
      private_key_file (str): the file path for the private_key.
        It can be a directory, in which case, the private key
        will be generated into that directoiry. When not
        provided a file is generated in the local directory.
      public_key_file (str): the file path for the public_key
        (see private_key_file).

    """

    ## Derive directory from the provided files (when possible).
    try:
      if os.path.isfile( private_key_file ) :
        conf_dir = os.path.dirname( private_key_file )
      elif os.path.isdir( private_key_file ) :
        conf_dir = private_key_file
    except:
      try:
        if os.path.isfile( public_key_file ) :
          conf_dir = os.path.dirname( public_key_file )
        elif os.path.isdir( public_key_file ) :
          conf_dir = public_key_file
      except:
        conf_dir = './'
    ## ensures that there is a private key
    try :
      private_key = self.load_private_key( private_key_file )
    except ConfigurationError as e:
      print( e.message )
      print( "WARNING: Generating new keys")
      private_key, public_key = self.generate_keys( sig_scheme )
      private_key_file, public_key_file = self.store_keys( private_key, key_format, conf_dir )
      print( f"  - private_file: {private_key_file}" )
      print( f"  - public_file: {public_key_file}" )
      ## updating the configuration (only necessary when directory is provided)
      self.conf[ ( 'tls13', 'v1' )  ] ['private_key'] = private_key_file
      self.conf[ ( 'tls13', 'v1' )  ] ['public_key'] = [ public_key_file ]

#    finally:
    private_key = self.load_private_key( private_key_file )
    sig_scheme = self.key_algo( private_key )

    ## ensure there is a corresponding public key or certificate
    ## stored in the file and that public key matches the private key
    try :
      ## taking the last file mentioned in the public_key list
      public_key_file = self.conf[ ( 'tls13', 'v1' )  ] ['public_key'][ -1 ]
      public_key, cert_type = self.load_public_key( public_key_file )
      if self.key_algo( public_key ) !=  sig_scheme:
        raise ConfigurationError( f"public ({self.key_algo( public_key )}) and "\
                f"private keys ({self.key_algo( private_key )}) do not match" )
    except:
      private_key = self.load_private_key( private_key_file )
      public_key = private_key.public_key( )
##      self.store_keys( public_key, key_format, conf_dir )
      ##  we do store the private key as 1) storing the private key
      ## results in storing the public key. but also 2) storing the public
      ## key as a self-signed certificate requires the private key.
      self.store_keys( private_key, key_format, conf_dir )
    return private_key_file, public_key_file


  def set_connectivity( self, **kwargs ):
    """configure the connectivity informations """
    self.conf[ 'connectivity' ] = kwargs

  def set_tls13_debug( self, **kwargs ):
    """ updating the debug configuration

    There is a special check for the trace key
    """

    if 'trace' not in kwargs.keys() :
      kwargs[ 'trace' ] = False
    self.conf[ ( 'tls13', 'v1' ) ][ 'debug' ] = kwargs

  def set_tls13_authorization_type( self ):
    role = self.conf[ ( 'tls13', 'v1' ) ][ 'role' ]
    for k in list( self.conf[ ( 'tls13', 'v1' ) ][ 'type_authorized'] ) :

      if ( role == 'client' and k[0:2] == 's_' ) or\
         ( role == 'server' and k[0:2] == 'c_' ):
        self.conf[ ( 'tls13', 'v1' ) ][ 'type_authorized'].remove( k )

  def set_tls13_cs_signing_key( self ):
    """configure the CS signing keys and associated internal variables"""

    try:
      private_key_file = self.conf[ ( 'tls13', 'v1' ) ][ 'private_key' ]
    except KeyError:
      private_key_file = None
    try:
      public_key_file = self.conf[ ( 'tls13', 'v1' )  ] ['public_key'][ -1 ]
    except KeyError:
      public_key_file = None
    try:
      sig_scheme = self.conf[ ( 'tls13', 'v1' )  ] [ 'sig_scheme' ][ 0 ]
    except KeyError:
      sig_scheme = 'ed25519'
#    print( f"private_key_file: {private_key_file}" )
#    print( f"public_key_file: {public_key_file}" )
#    print( f"conf: {self.conf}" )
    private_key_file, public_key_file = self.set_tls13_keys( \
                                          private_key_file=private_key_file, \
                                          public_key_file=public_key_file, \
                                          sig_scheme=sig_scheme,\
                                          key_format='X509' )
    ## Once the files have been checked are correct we can load them
    ## we are just checking loading is performed correctly
    public_key, cert_type = self.load_public_key( public_key_file )
    private_key = self.load_private_key( private_key_file )

    ## updating self.conf
    self.conf[ ( 'tls13', 'v1' ) ][ 'private_key' ] = private_key_file
    ## handling the public key file is a bit more complex as the
    ## file is a file list when key format X509 is used.
    public_key_conf_status = False
    if 'public_key' in self.conf[ ( 'tls13', 'v1' )  ].keys( ) :
      if isinstance( self.conf[ ( 'tls13', 'v1' )  ]['public_key'], list ):
        if self.conf[ ( 'tls13', 'v1' )  ] ['public_key'][ -1 ] == public_key_file :
          public_key_conf_status = True
    if public_key_conf_status is False:
      self.conf[ ( 'tls13', 'v1' )  ] ['public_key'] = [ public_key_file ]
    self.conf[ ( 'tls13', 'v1' ) ] [ '_private_key' ] = private_key
    self.set_tls13_cs_public_signing_key( )

  def set_tls13_cs_public_signing_key( self ):
    """ set the configuration related to the public key

    The reason we split the private and public key is to enable the
    TLS Engine to configure the necessary (public) parameteres to build
    its Certificate payload or identify the signature algorithm and keys.
    """

    public_key_file = self.conf[ ( 'tls13', 'v1' )  ] ['public_key'][ -1 ]
    public_key, cert_type = self.load_public_key( public_key_file )
    self.conf[ ( 'tls13', 'v1' ) ] [ '_public_key' ] = public_key
    self.conf[ ( 'tls13', 'v1' ) ] [ '_cert_type' ] = cert_type
    cert_entry_list = self.load_cert_entry_list()
    self.conf[ ( 'tls13', 'v1' ) ][ '_cert_entry_list' ] = cert_entry_list
    finger_print_dict = {}
    finger_print_entry_list = []
    for cert_entry in cert_entry_list:
      public_bytes = cert_entry[ 'cert' ]
      digest = Hash( SHA256() )
      digest.update( public_bytes )
      finger_print = digest.finalize()[ :4 ]
      finger_print_dict[ finger_print ] = public_bytes
      finger_print_entry_list.append( { 'finger_print' : finger_print, 'extensions': [] } )
    self.conf[ ( 'tls13', 'v1' ) ] [ '_finger_print_entry_list' ] = finger_print_entry_list
    self.conf[ ( 'tls13', 'v1' ) ] [ '_finger_print_dict' ] = finger_print_dict


class CLI:
  """ Generates conf file from command line arguments.

  Template can be manually generated.
  The template is expected to provide a single port
  to each configuration.
  The template for a CS is expected to look this way:

  ..  code-block:: python

    cs_conf_template = {
    'connectivity' : {
       'type': 'tcp',
       'ip' : '127.0.0.1',
       'port' : 9402
      },
      ## logs are redirected to stdout especially when
      ## the cs is running in the enclave.
      'log' : None,
      ( 'tls13', 'v1' ) : {
        'public_key' : [ os.path.join( conf_dir, 
                         '_Ed25519PublicKey-ed25519-X509.der' ) ],
        'private_key': os.path.join( conf_dir, 
                         '_Ed25519PrivateKey-ed25519-pkcs8.der' ) ,
        'sig_scheme': ['ed25519']
      }
    }

  The reason we define a class is to be able to manage the
  various configuration of the CS with which also includes
  the implementation of illustrated_tls13 as well as the
  instantiation inside a sgx enclave.

  Attributes:
    connectivity: the type of connectivity for the CS. Possible values are
      'lib_cs' (default), 'tcp', 'persistent_tcp'
    debug: a boolean value that indicates if the debug mode is enabled
    test_vector_mode: indicates if values are expected to be taken from
      a test_vector file or recorded in a test_vector file or no 
      test_vector file is considered at all. The possible values are
      'check', 'record' or None (default).
    test_vector_file: when the test_vector_mode is not set to None, 
      the file the values are read from or written to. By default 
      it is set to None. 
    host: The IP address of the FQDN of the CS. This parameter is
      mostly useful for entities connecting to the CS as opposed to 
      the CS. The default is '127.0.0.1'.
    port: integer that defines the port of the CS set to 9400 by default.
    sig_scheme: the signature scheme the CS implements. This parameter
      may be redundant when the key is specified - but not necessarily.
      For example a RSA key may be used with different signature scheme.
      The scheme is also used to generate a key when the key is not 
      provided. By default, it is set to 'ed25519'.
    key: The file that contains the private key. By default it is set 
      to None and the key is generated at the instantiation of the CS.
    cert: The public key or certificate.
  """


  def __init__( self, connectivity:str='lib_cs',
                      debug:bool=False,
                      test_vector_mode=None,
                      test_vector_file=None,
                      host='127.0.0.1',
                      port=9400,
                      sig_scheme='ed25519',
                      key=None,
                      cert=None,
                      ):

    self.connectivity = connectivity
    self.debug = debug
    self.test_vector_mode = test_vector_mode
    self.test_vector_file = test_vector_file
    self.host = host
    self.port = port
    self.sig_scheme = sig_scheme
    self.key = key
    self.cert = cert


  def get_template( self ):
    """ generates the template

    Note that log is set to None and redirects the messages
    to the outputs.
    We define such value as to prevent writing an external file
    when teh server runs in an SGX enclave. We should define --log_level
    --log_file and force these values when SGX is enabled.
    """
    return { 'log' : None,
             'connectivity' : self.get_connectivity( ),
             ( 'tls13', 'v1' ) : self.get_tls13( ) }

  def get_debug( self ):
    debug_template = { 'trace' : self.debug }
    if self.test_vector_mode not in [ 'check', 'mode', None ]:
      raise ValueError( f"Invalid test_vector_mode value {self.test_vector_mode}")
    if self.test_vector_file is None:
      if self.test_vector_mode == 'check' :
        raise ValueError( f"non coherent values for test_vector_file"\
              f"{self.test_vector_file} and test_vector_mode {self.test_vector_mode}" )
      elif self.test_vector_mode == 'record' :
        self.test_vector_file = "./test_vector_file.json"
    else:
      if self.test_vector_mode is None:
        self.test_vector_mode == 'check'

    if self.test_vector_mode is not None:
      debug_template[ 'test_vector' ] = {}
      debug_template[ 'test_vector' ][ 'file' ][ self.test_vector_file ]
      debug_template[ 'test_vector' ][ 'mode' ][ self.test_vector_mode ]
    return debug_template


  def get_connectivity( self ):
    if self.connectivity not in [ 'lib_cs', 'tcp', 'persistent_tcp' ]:
      raise ValueError ( f"connectivity ({self.connectivity}) MUST be in "\
            f"'lib_cs', 'tcp', 'persistent_tcp'" )
    connectivity_template = { 'type' : self.connectivity }
    if connectivity_template [ 'type' ] != 'lib_cs' :
      connectivity_template [ 'ip' ] = self.host
      connectivity_template [ 'port' ] = self.port
    return connectivity_template

  def get_tls13( self ):
    tls13_template = { 'sig_scheme': [ self.sig_scheme ] }
    tls13_template[ 'public_key' ] = [ self.cert ]
    tls13_template[ 'private_key' ] = self.key
    tls13_template[ 'debug' ] = self.get_debug( )
    return tls13_template


  def get_parser( self, env:bool=True, conf_dir:str='./',
                  parser=None):
    """ This function returns a parser to start the CS

    The CS can be started as a regular library in which case
    only library related parameters are provided.
    On the other hand, the CS MAY also requires some speciifc
    OS configuration to start the service.
    These parameters are not handled by the CS library itself,
    but actually defines how the library is started.

    Args:
      env (bool) when set to False indicates that only the
        library parameters are provided.
        When set to True, this includes OS specific environement
        configuration parameters.
      conf_dir (str): The path to the CS directory. It is
        expected to contain the CS enclave as well as some
        parameters such as the keys, certificate.

    """


    if parser is None:
      if env is True:
        description = \
        """
        This scripts launches the Crypto Service in various modes.
        These modes includes:

        1) launching the Crypto Service in the rich environement
        - that is like a standard python library in the OS.
        This is the defaul mode.
        
        2) lauching the Crypto Service in an SGX enclave using
        Gramine. This is indicated by the -sgx or
        --gramine_sgx option.
        
        3) launching the Crypto Service with Gramine but NOT
        in a SGX enclave.
        This is indicated by the -g or --grammine_direct option

        To start the Crypto Service using SGX, the Crypto Service
        the enclave MUST have been previously built.
        Building the enclave is performed using the -b or
        --gramine_build option.

        This script can be seen as setting the expected environement
        to start the Crypto Service.
        However the Crypto Service (with its expected configuration)
        is actually started by the start_cs.py script.
        Most of the arguments are passed to that start_cs.py

        Example:
          ## Building the enclave (needs only to be performed once)
          ./crypto_service --gramine_build

          ## Starting the CS in an SGX enclave
          ./crypto_service --connectivity tcp --sig_scheme ed25519
            --gramine_sgx
        """

      else :
        description = \
        """
        This script configure the CS but not the OS environment
        parameters.
        """
      parser = argparse.ArgumentParser( description=description )

##    parser = argparse.ArgumentParser( description=description )
    parser.add_argument( '-con', '--connectivity', type=ascii, \
      default='tcp', nargs='?', \
      help='Crypto Service  connectivity [ tcp, persistent_tcp ]')
    parser.add_argument( '-host', '--host', type=ascii, \
      default='127.0.0.1', nargs='?', \
      help='Crypto Service  IP address or hostname')
    parser.add_argument( '-port', '--port', type=int, \
      default='9400', nargs='?', \
      help='Crypto Service  port')
    parser.add_argument( '-sig', '--sig_scheme', \
      type=ascii, default='ed25519', nargs='?', \
      help='Crypto Service  signature scheme  [ ed25519 ]')
    ## We mandate the CS to have a public / private key
    ## and currently do not consider the key not to be used.
    ## this may not represent the case of an unauthenticated
    ## TLS client.
    ##
    ## With SGX the file needs to be local - as it is configured
    ## in the  python template.
    ## There might need to take different path depending on
    ## whether crypto_service or strat_cs is used.
    ## currently We assume that one is in the local directory.
#    key_file = os.path.join( './sig_key_dir', \
#                 '_Ed25519PrivateKey-ed25519-pkcs8.der' )
#    key_file = os.path.join( conf_dir, 'sig_key_dir', \
#                 '_Ed25519PrivateKey-ed25519-pkcs8.der' )
    parser.add_argument( '-key', '--key', \
      type=pathlib.Path, default=None, nargs='?', \
      help='Crypto Service  private key')
#    cert_file = os.path.join( conf_dir, 'sig_key_dir', \
#                  '_Ed25519PublicKey-ed25519-X509.der' )
#    cert_file = os.path.join(  './sig_key_dir', \
#                  '_Ed25519PublicKey-ed25519-X509.der' )
    parser.add_argument( '-cert', '--cert', \
      type=pathlib.Path, default=None, nargs='?', \
      help='Crypto Service  public key')
    parser.add_argument( '-debug', '--debug', default=False,  \
      action='store_const', const=True, \
      help='Crypto Service debug_mode')
    parser.add_argument( '-tv_mode', '--test_vector_mode', \
      type=ascii, default=None, nargs='?', \
      help="Crypto Service  test vector mode [ 'check', 'record', None ]" )
    parser.add_argument( '-tv_file', '--test_vector_file', \
      type=pathlib.Path, default=None, nargs='?', \
      help='Crypto Service  test vector file')
#    print( f"env - 2: {env}" )
    if env is True:
      parser.add_argument( '-sgx', '--gramine_sgx', default=False,  \
        action='store_const', const=True, \
        help='Crypto Service is run into SGX (gramine)')
      parser.add_argument( '-g', '--gramine_direct', default=False,  \
        action='store_const', const=True, \
        help='Crypto Service is run into SGX (gramine)')
      parser.add_argument( '-b', '--gramine_build', default=False,  \
        action='store_const', const=True, \
        help='Build the Crypto Service into the enclave')
      parser.add_argument( '-sec_prov', '--secret_provisioning',\
        default=False,  action='store_const', const=True, \
        help='Enable Secret Provisioning')
      parser.add_argument( '-ra_type', '--ra_type', type=ascii, \
        default='None', nargs='?', help='type of remote attestation')
      parser.add_argument( '-ra_spid', '--ra_spid', type=ascii, \
        default='None', nargs='?', \
        help='Service Provider ID for remote attestation ')
      parser.add_argument( '-ra_linkable', '--ra_linkable', \
        type=ascii, default='None', nargs='?', \
        help='Specify whether EPID attestation is linkable or not.')
      parser.add_argument( '-gramine_dir', '--gramine_dir',\
        type=ascii, default='None', nargs='?', \
        help='Specify whether EPID attestation is linkable or not.')

    return parser

  def init_from_args( self, args ):
    """initializes the variables from the command lines arguments

    The command lines have been parsed with the parser obtained
    from get_parser.

    Args:
      args: the output returned by parser.parse_args( )
    """
    if args.test_vector_mode is None:
      test_vector_mode = None
    else:
      test_vector_mode = args.test_vector_mode[ 1:-1 ]
    if args.test_vector_file is None:
      test_vector_file = None
    else:
      test_vector_file = args.test_vector_file[ 1:-1 ]

    self.connectivity = args.connectivity[1:-1]
    self.debug = args.debug
    self.test_vector_mode=test_vector_mode
    self.test_vector_file=test_vector_file
    self.host=args.host[1:-1]
    self.port=args.port
    self.sig_scheme=args.sig_scheme[1:-1]
    self.key=args.key
    self.cert=args.cert

  def copy_and_update_file_path( self, origin_file_path, \
      GRAMINE_DIR, sub_dir='' ):
    """ copy and upadte the file paths so it can be used by gramine

    Gramine requires all elements it trusts to be below the
    script starting Gramine.

    When a key file is provided, the file is copy under the
    specific con_dir location and the new file path is returned.
    The new file path is relative to the gramine directory.
    """
    try:
      shutil.copy( origin_file_path, os.path.join( GRAMINE_DIR, sub_dir ) )
    except shutil.SameFileError :
      pass
    return os.path.join( './', sub_dir, \
             os.path.basename( origin_file_path) )





from _io import BytesIO
import traceback
from copy import deepcopy
from construct.core import *
from construct.lib import *
import json

def title(title):
    """ print title in a square box

    To enhance the readability of the tests, this function prints in the
    terminal teh string title in a square box.

    Args:
        title (str): the string
    """
    space = "    "
    title = space + title + space
    h_line = '+'
    for character in title:
        h_line += '-'
    h_line += '+\n'
    print('\n' + h_line + '|' + title + '|\n' + h_line )


def compare( data_struct1, data_struct2):
  """ compares two data structures """
  if isinstance( data_struct1, (dict, Container) ) and\
     isinstance( data_struct2, (dict, Container) ) :
    ## removing unsignificant variable for container, i.e. used for teh
    ## purpose of data processing
    data_keys = []
    for data_struct in [ data_struct1, data_struct2]:
      keys = list(data_struct.keys())
      if isinstance(data_struct, Container):
        for k in keys[:]:
          if k[0] == '_' or k == 'reserved':
            keys.remove(k)
      data_keys.append(set(keys))
    ## comparing keys
    if not data_keys[0] == data_keys[1]:
      k1 = data_keys[0]
      k2 = data_keys[1]
      raise Exception(\
        "\n    - k1: %s"%k1 + "\n    - k2: %s"%k2 +\
        "\n    - keys in k1 not in k2: %s"%k1.difference(k2) +\
        "\n    - keys in k2 not in k1 :%s"%k2.difference(k1) +\
        "\n    - data_struct1: %s"%data_struct1 +\
        "\n    - data_struct2: %s"%data_struct2 )
    for k in data_keys[1] :
      compare(data_struct1[k], data_struct2[k])
  elif isinstance( data_struct1, (list, set, ListContainer)) and\
     isinstance( data_struct2, (list, set, ListContainer)) :
    for i in range(len(data_struct1)):
      if ( None in data_struct1 and None not in data_struct2):
        data_struct1.remove(None)
      if ( None in data_struct2 and None not in data_struct1):
        data_struct2.remove(None)

    if len(data_struct1) == len(data_struct2):
      for i in range(len(data_struct1)):
         compare(data_struct1[i], data_struct2[i])
    else:
      raise Exception( f"lenght do not match" \
        f"\n    - data_struct1 [{type(data_struct1)}] [len: {len(data_struct1)}]: {data_struct1}"\
        f"\n    - data_struct1 [{type(data_struct2)}] [len: {len(data_struct2)}]: {data_struct2}" )
  elif isinstance( data_struct1, (str, EnumIntegerString)) and\
     isinstance( data_struct2, (str, EnumIntegerString)) :
    if str(data_struct1) != (data_struct2):
      raise Exception( \
        "\n    - data_struct1 [%s] : %s"%(type(data_struct1), data_struct1) +\
        "\n    - data_struct2 [%s] : %s"%(type(data_struct2), data_struct2) )
  else:
    if data_struct1 != data_struct2:
      if data_struct1 in  [ None, b'' ] and data_struct2 in [ None, b'' ]:
        pass
      else:
         raise Exception( \
           "\n    - data_struct1 [%s] : %s"%(type(data_struct1), data_struct1) +\
           "\n    - data_struct2 [%s] : %s"%(type(data_struct2), data_struct2) )


def obj2json( data:dict ) -> dict:
  """ converts a bytes or bytearray value to string """
  json_data = deepcopy( data )
  if isinstance( json_data, BytesIO ) :
    json_data = json_data.read()
#  if isinstance( json_data, bytes ) or\
  if type( json_data ) is bytes or\
     isinstance( json_data, bytearray ) : #or\
#     isinstance( json_data, BytesIO) :
    json_data = json_data.hex( '-' )
  elif isinstance( json_data, dict ):
    for k in json_data.keys():
##      if isinstance( json_data[ k ], BytesIO ):
##        del json_data[ k ]
##      else:
        json_data[ k ] = obj2json( json_data[ k ] )
  elif isinstance( json_data, list ):
    for i in range( len( json_data ) ):
      json_data[ i ] = obj2json( json_data[ i ] )
  else:
    c = json_data.__class__.__name__
##    if c == 'BytesIO':
##      print( "class: %s, object: %s, type: %s"%(c, json_data.read(), type( json_data ) ) )

  return json_data


def test_struct( struct, data_struct, ctx_struct={}, \
                 ext_title='', no_title=False, \
                 io_check=True, print_data_struct=False, \
                 print_binary=False, print_data=True ):
  """ compares a data structure ( i.e. a dictionary to the structure itself  """
#  print( f" --- input : {struct} - {data_struct}" )
  binary = struct.build(data_struct, **ctx_struct)
#  print( f" --- binary  : {binary}" )
  data = struct.parse(binary, **ctx_struct)
#  print( f" --- data  : {data}" )

  if not no_title:
    try:
      name = data._name
    except(AttributeError):
      name = ''
    title("%s [%s]  structure"%(name, ext_title))

  if print_data_struct == True:
    print("struct: %s"%json.dumps( obj2json( data_struct ), indent=2) )
  if print_binary == True:
    print("bytes: %s"%binary.hex( '-' ) )
  if print_data == True:
    print("struct: %s"%data)
  if io_check:
    try:
      #print( f"{data_struct} - {data}" )
      compare( data_struct, data )
    except AssertionError as e:
      _, _, tb = sys.exc_info()
      traceback.print_tb(tb) # Fixed format
      tb_info = traceback.extract_tb(tb)
      filename, line, func, text = tb_info[-1]
      print('An error occurred on line {} in statement {}'.format(line, text))
      print(e)
      exit(1)
  return binary, data


def match( req, resp ):
  """ tests if the re and resp match """
  if req[ 'designation' ] != resp[ 'designation' ]:
    raise ValueError( f"designation does not match {req} / {resp}" )
  if req[ 'version' ] != resp[ 'version' ]:
    raise ValueError( f"version does not match {req} / {resp}" )
  if req[ 'id' ] != resp[ 'id' ]:
    raise ValueError( f"id does not match {req} / {resp}" )





from construct.core import *
from construct.lib import *
from construct.debug import *
#import sys
#sys.path.insert(0, '/home/emigdan/gitlab/pytls13/src/pytls13') #pytls13
import pytls13.struct_tls13 as tls

""" TLS 1.3 related structures modified by lurk"""

## Extensions

## supported version
#SupportedVersions = Struct(
#          select (Handshake.msg_type) {
#              case client_hello:
#                   ProtocolVersion versions<2..254>;
#
#              case server_hello: /* and HelloRetryRequest */
#                   ProtocolVersion selected_version;
#          };
#)

## 

##  pytls13
#-#-# signature_algorithm 
#-#-# SignatureScheme = Enum( Bytes(2), 
#-#-#  rsa_pkcs1_sha256 = b'\x04\x01',
#-#-#  rsa_pkcs1_sha384 = b'\x05\x01',
#-#-#  rsa_pkcs1_sha512 = b'\x06\x01',
#-#-#  ecdsa_secp256r1_sha256 = b'\x04\x03',
#-#-#  ecdsa_secp384r1_sha384 = b'\x05\x03',
#-#-#  ecdsa_secp521r1_sha512 = b'\x06\x03',
#-#-#  rsa_pss_rsae_sha256 = b'\x08\x04',
#-#-#  rsa_pss_rsae_sha384 = b'\x08\x05',
#-#-#  rsa_pss_rsae_sha512 = b'\x08\x06',
#-#-#  ed25519 = b'\x08\x07',
#-#-#  ed448 = b'\x08\x08',
#-#-#  rsa_pss_pss_sha256 = b'\x08\x09',
#-#-#  rsa_pss_pss_sha384 = b'\x08\x0a',
#-#-#  rsa_pss_pss_sha512 = b'\x08\x0b',
#-#-#  rsa_pkcs1_sha1 = b'\x02\x01',
#-#-#  ecdsa_sha1 = b'\x02\x03',
#-#-#)
#-#-#
#-#-#SignatureSchemeList = Struct(
#-#-#  'supported_signature_algorithms' / Prefixed(BytesInteger(2),\
#-#-#                                       GreedyRange(SignatureScheme))
#-#-#)
#-#-#
#-#-### psk_key_exchange_modes
#-#-#
#-#-#PskKeyExchangeMode = Enum(BytesInteger(1), 
#-#-#  psk_ke = 0,
#-#-#  psk_dhe_ke = 1,
#-#-#)
#-#-#
#-#-#PskKeyExchangeModes = Struct(
#-#-#  'ke_modes' / Prefixed(BytesInteger(1), GreedyRange(PskKeyExchangeMode))
#-#-#)
#-#-#
#-#-### post_handshake_auth
#-#-#
#-#-#PostHandshakeAuth = Struct()
#-#-#
#-#-### supported_group 
#-#-#
#-#-#NamedGroup = Enum( Bytes(2),
#-#-###  unallocated_RESERVED(0x0000),
#-#-###  /* Elliptic Curve Groups (ECDHE) */
#-#-###  obsolete_RESERVED(0x0001..0x0016),
#-#-#  secp256r1 = b'\x00\x17', 
#-#-#  secp384r1 = b'\x00\x18', 
#-#-#  secp521r1 = b'\x00\x19',
#-#-###  obsolete_RESERVED(0x001A..0x001C',
#-#-#  x25519 = b'\x00\x1D', 
#-#-#  x448 = b'\x00\x1E',
#-#-###  /* Finite Field Groups (DHE) */
#-#-#  ffdhe2048 = b'\x01\x00', 
#-#-#  ffdhe3072 = b'\x01\x01', 
#-#-#  ffdhe4096 = b'\x01\x02',
#-#-#  ffdhe6144 = b'\x01\x03', 
#-#-#  ffdhe8192 = b'\x01\x04',
#-#-###  /* Reserved Code Points */
#-#-###  ffdhe_private_use(0x01FC..0x01FF),
#-#-###  ecdhe_private_use(0xFE00..0xFEFF),
#-#-###  obsolete_RESERVED(0xFF01..0xFF02), (0xFFFF)
#-#-#)
#-#-#
#-#-#NamedGroupList = Struct(
#-#-#  'named_group_list' / Prefixed(BytesInteger(2), GreedyRange(NamedGroup))
#-#-#)
#-#-#
#-#-#
#-#-### Key Share
#-#-#
#-#-#KeyShareEntry = Struct(
#-#-# '_name' / Computed('KeyShareEntry'),
#-#-#  'group' / NamedGroup, 
#-#-#  'key_exchange' / Prefixed(BytesInteger(2), Switch(this.group,
#-#-#  { 'secp256r1' : Struct(
#-#-#      'legacy_form' / Const(4, BytesInteger(1)), 
#-#-#      'x' / BytesInteger(32),
#-#-#      'y' / BytesInteger(32)
#-#-#      ), 
#-#-#    'secp384r1' : Struct(
#-#-#      'legacy_form' / Const(4, BytesInteger(1)), 
#-#-#      'x' / BytesInteger(48),
#-#-#      'y' / BytesInteger(48)
#-#-#      ), 
#-#-#    'secp521r1' : Struct(
#-#-#      'legacy_form' / Const(4, BytesInteger(1)), 
#-#-#      'x' / BytesInteger(66),
#-#-#      'y' / BytesInteger(66)
#-#-#      ), 
#-#-#    'x25519' : Bytes(32),
#-#-#    'x448' : Bytes(56),
#-#-#  }))
#-#-#)

### only for LURK
EmptyKeyShareEntry = Struct(
 '_name' / Computed('EmptyKeyShareEntry'),
  'group' / tls.NamedGroup,
  'key_exchange' / Prefixed( BytesInteger(2), Const(b'') )
)

#-#-#KeyShareClientHello = Struct(
#-#-# '_name' / Computed('KeyShareClientHello'),
#-#-# 'client_shares' / Prefixed(BytesInteger(2), GreedyRange(KeyShareEntry)) 
#-#-#)

## only for LURK
PartialKeyShareClientHello = Struct(
 '_name' / Computed('KeyShareClientHelloEmpty'),
## 'client_shares' / Prefixed(BytesInteger(2), GreedyRange( Select( tls.KeyShareEntry, EmptyKeyShareEntry ) ) ) 
 'client_shares' / Prefixed(BytesInteger(2), GreedyRange( Select( EmptyKeyShareEntry, tls.KeyShareEntry ) ) ) 
)

## only for LURK
#KeyShareClient = Select( 
#  KeyShareClientHello, KeyShareClientHelloEmpty
#)

#-#-#KeyShareHelloRetryRequest = Struct(
#-#-# '_name' / Computed('KeyShareHelloRetryRequest'),
#-#-#  'selected_group' / NamedGroup
#-#-#)
#-#-#
#-#-#KeyShareServerHello = Struct(
#-#-# '_name' / Computed('KeyShareServerHello'),
#-#-#  'server_share' / KeyShareEntry
#-#-#)

## defining the EmptyKeyShareEntry as a potential structure
## for KeyShareEntry.


KeyShareServerHelloEmpty = Struct(
 '_name' / Computed('KeyShareServerHelloEmpty'),
  'server_share' / EmptyKeyShareEntry
)

## overwritting a TLS 1.3 structure
## LURK and non LURK
KeyShareServer = Select(
  tls.KeyShareServerHello, KeyShareServerHelloEmpty,
  tls.KeyShareHelloRetryRequest
)

#-#-### pre_shared_key
#-#-#PskIdentity = Struct(
#-#-#  'identity' / Prefixed( BytesInteger(2), GreedyBytes),
#-#-#  'obfuscated_ticket_age' / Bytes(4)
#-#-#)
#-#-#
#-#-#PskBinderEntry = Struct(
#-#-#  'binder' / Prefixed(BytesInteger(1), GreedyBytes)
#-#-#)
#-#-#
#-#-#OfferedPsks = Struct(
#-#-#  'identities' / Prefixed(BytesInteger(2), GreedyRange(PskIdentity)), 
#-#-#  'binders' / Prefixed(BytesInteger(2), GreedyRange(PskBinderEntry))
#-#-#)


## Only for LURK

#ExtendedPskIdentity = Struct(
#  '_name' / Computed(''), 
#  'identity' / Prefixed( BytesInteger(2), GreedyBytes ),
#  'obfuscated_ticket_age' / Bytes(4)
#  'tls_hahs' / TLSHash, 
#  'psk_bytes' / Prefixed( BytesInteger(2), GreedyBytes ),
#)

OfferedPsksWithNoBinders = Struct(
  'identities' / Prefixed(BytesInteger(2), GreedyRange(tls.PskIdentity)) 
)



## server_certificate

## https://www.iana.org/assignments/tls-extensiontype-values/tls-extensiontype-values.xhtml#tls-extensiontype-values-3
#-#-#CertificateType = Enum( BytesInteger(1), 
#-#-#  X509 = 0, 
#-#-#  RawPublicKey = 2
#-#-#)

##   struct {
##           select(ClientOrServerExtension) {
##               case client:
##                 CertificateType client_certificate_types<1..2^8-1>;
##               case server:
##                 CertificateType client_certificate_type;
##           }
##   } ClientCertTypeExtension;
##
##   struct {
##           select(ClientOrServerExtension) {
##               case client:
##                 CertificateType server_certificate_types<1..2^8-1>;
##               case server:
##                 CertificateType server_certificate_type;
##           }
##   } ServerCertTypeExtension;


# supported_versions

##ProtocolVersion = Bytes( 2 )
##
##SupportedVersions = Struct( Switch( this.extension_type, 
##  {
##    'client_hello' : Prefixed( BytesInteger(1), GreedyRange( ProtocolVersion ))
##    'server_hello' : ProtocolVersion
##
##  }, default=Bytes )
##)
##
##struct {
##          select (Handshake.msg_type) {
##              case client_hello:
##                   ProtocolVersion versions<2..254>;
##
##              case server_hello: /* and HelloRetryRequest */
##                   ProtocolVersion selected_version;
##          };
##      } SupportedVersions

## Extension structure

#-#-#ExtensionType = Enum( BytesInteger(2), 
#-#-#  server_name = 0,                              
#-#-#  max_fragment_length = 1,                      
#-#-#  status_request = 5,                           
#-#-#  supported_groups = 10,                        
#-#-#  signature_algorithms = 13,                    
#-#-#  use_srtp = 14,                                
#-#-#  heartbeat = 15,                               
#-#-#  application_layer_protocol_negotiation = 16,  
#-#-#  signed_certificate_timestamp = 18,            
#-#-#  client_certificate_type = 19,                 
#-#-#  server_certificate_type = 20,                 
#-#-#  padding = 21,                                 
#-#-#  pre_shared_key = 41,                          
#-#-#  early_data = 42,                              
#-#-#  supported_versions = 43,                      
#-#-#  cookie = 44,                                  
#-#-#  psk_key_exchange_modes = 45,                  
#-#-#  certificate_authorities = 47,                 
#-#-#  oid_filters = 48,                             
#-#-#  post_handshake_auth = 49,                     
#-#-#  signature_algorithms_cert = 50,               
#-#-#  key_share = 51,                               
#-#-#)
#-#-#
#-#-### To do list all possible extensions with a Switch
#-#-### The default is set to Bytes, This is expected to be usefull for
#-#-### parsing but cannot be used for building when a structure is provided.
#-#-### Instead the rwa data may be provided - to be tested. 
#-#-#
###Extension = Struct(
###  '_name' / Computed('Extension'),
###  'extension_type' / tls.ExtensionType,
###  'extension_data' /  Prefixed(BytesInteger(2),
###                      Switch(this.extension_type,
###    {
###      'supported_groups' : tls.NamedGroupList,
###      'signature_algorithms' : tls.SignatureSchemeList, 
###      'pre_shared_key' : Switch(this._._msg_type,
###         {
###           'client_hello' : tls.OfferedPsks, 
###           'server_hello' : BytesInteger( 2 ) 
###         }
###        ),
###      'post_handshake_auth' : tls.PostHandshakeAuth,
###      'psk_key_exchange_modes' : tls.PskKeyExchangeModes, 
###      'key_share': Switch(this._._msg_type, 
###         {
###          'client_hello' : tls.KeyShareClientHello, 
###          'server_hello' : tls.KeyShareServer, 
###         }
###        )
###    }, default=Bytes)
###  ) 
###)

##LURK only -- we may do the same with ServerHello
## PartialCHExtension

###PartialCHExtension = Struct(
###  '_name' / Computed('PartialCHExtension'),
###  'extension_type' / tls.ExtensionType,
###  'extension_data' /  Prefixed(BytesInteger(2),
###                      Switch(this.extension_type,
###    {
######      'supported_groups' : tls.NamedGroupList,
######      'signature_algorithms' : tls.SignatureSchemeList, 
###      'pre_shared_key' : Switch(this._._msg_type,
###         {
###           'client_hello' : OfferedPsksWithNoBinders, 
###           'server_hello' : BytesInteger( 2 ) 
###         }
###        ),
######      'post_handshake_auth' : tls.PostHandshakeAuth,
######      'psk_key_exchange_modes' : tls.PskKeyExchangeModes, 
###      'key_share': Switch(this._._msg_type, 
###         {
###          'client_hello' : Select( PartialKeyShareClientHello, tls.KeyShareClientHello ),
###          'server_hello' : tls.KeyShareServer, 
###         }
###        )
#####    }, default=Bytes)
###    }, default=StreamError )
###  ) 
###)

## Note: We wanted to be able to say Select ( PartialCHExtension, Extension ).  The expected behavior would be that if a building error happens in the PartialCHExtension, then structure from the Extension would be considered. 

## The restriction of the PartialCHExtensionType prevents an object with an unknown extnesion to be generated. This is expected to make Select work as expected. 
## However, to parse it, Enum does not raise any issue when it does not find the corresponding string. As a result parsing does not generates any error at that point. 
## Switch may generate that error, bu an empty object is returned when the key does not appear. This result sin PartialCHExtension scrapping the extension not mentioned explicitly. 
## We currenlty do not rely on Select and provide an explicit list of extensions.
## We need to handle for build and parse unknown extensions. This woudl at least enable us to only care about those that are necessary to parse by the CS
## --> If then Else may try all possibilities and define what to do at parsing. 

## We also need to check how a Enum could return an error that triggers Select to look at the next structure. 

## these are the types of extensions whose structure is modified by LURK
PartialCHExtensionType = Enum( BytesInteger(2), 
  pre_shared_key = 41,
  key_share = 51,

) 

PartialCHExtension = Struct(
  '_name' / Computed('PartialCHExtension'),
##  'extension_type' / PartialCHExtensionType,
  'extension_type' / tls.ExtensionType,
#  Probe( this.extension_type ),
  'extension_data' /  Prefixed(BytesInteger(2),
                      Switch(this.extension_type,
    {
      'pre_shared_key' : OfferedPsksWithNoBinders,
      'key_share' : Select( PartialKeyShareClientHello, tls.KeyShareClientHello ),
      ## repeated from Extension
      'server_name' : tls.ServerNameList,
      'supported_groups' : tls.NamedGroupList,
      'ec_point_format' : tls.ECPointFormatList,
      'signature_algorithms' : tls.SignatureSchemeList,
      'encrypt_then_mac' : Const( b'' ),
      'extended_master_secret' : Const( b'' ),
      'session_ticket' : GreedyBytes,
      'supported_versions' : tls.ClientSupportedVersions,
      'post_handshake_auth' : tls.PostHandshakeAuth,
      'psk_key_exchange_modes' : tls.PskKeyExchangeModes,
    } )
  ) 
)
## default indicates the error when no match is found for the 'extension_types'. This is the error raises when subcons does not match whihc is useful for Select. 


#-#-### ClientHello
#-#-#
#-#-#CipherSuite = Enum( Bytes(2),
#-#-#   TLS_AES_128_GCM_SHA256 = b'\x13\x01',
#-#-#   TLS_AES_256_GCM_SHA384 = b'\x13\x02',
#-#-#   TLS_CHACHA20_POLY1305_SHA256 = b'\x13\x03',
#-#-#   TLS_AES_128_CCM_SHA256 = b'\x13\x04',
#-#-#   TLS_AES_128_CCM_8_SHA256 = b'\x13\x05'
#-#-#)
#-#-#
ClientHello = Struct( 
  '_name' / Computed('ClientHello'),
  '_msg_type' / Computed('client_hello'),
  'legacy_version' / Const(b'\x03\x03'),
  'random' / Bytes(32),
  'legacy_session_id' / Prefixed(BytesInteger(1), GreedyBytes ),
  'cipher_suites' / Prefixed(BytesInteger(2), GreedyRange(tls.CipherSuite)),
  'legacy_compression_methods' / Prefixed(BytesInteger(1), GreedyBytes),
##  'extensions' / Prefixed(BytesInteger(2), GreedyRange( Extension))
##  'extensions' / Prefixed(BytesInteger(2), GreedyRange( Select( tls.Extension, PartialCHExtension ) ) )
  'extensions' / Prefixed(BytesInteger(2), GreedyRange( PartialCHExtension ) )
)


## LURK only
###PartialClientHello = Struct( 
###  '_name' / Computed('ClientHello'),
###  '_msg_type' / Computed('client_hello'),
###  'legacy_version' / Const(b'\x03\x03'),
###  'random' / Bytes(32), 
###  'legacy_session_id' / Prefixed(BytesInteger(1), GreedyBytes ),
###  'cipher_suites' / Prefixed(BytesInteger(2), GreedyRange(tls.CipherSuite)),
###  'legacy_compression_methods' / Prefixed(BytesInteger(1), GreedyBytes),
###  'extensions' / Prefixed(BytesInteger(2), GreedyRange(PartialCHExtension))
###)


#-#-### ServerHello
#-#-#
ServerHello = Struct(
  '_name' / Computed('ServerHello'),
  '_msg_type' / Computed('server_hello'),
  'legacy_version' / Const(b'\x03\x03'),
  'random' / Bytes(32),
  'legacy_session_id_echo' / Prefixed(BytesInteger(1), GreedyBytes),
  'cipher_suite' / tls.CipherSuite,
  'legacy_compression_method' / Const(b'\x00'),
  'extensions' / Prefixed(BytesInteger(2), GreedyRange(tls.Extension))
)
#-#-#
#-#-#
#-#-### NewsessionTicket
#-#-#NewSessionTicket = Struct(
#-#-#  '_name' / Computed('NewTicketSession'),
#-#-#  '_msg_type' / Computed('new_ticket_session'),
#-#-#  'ticket_lifetime' / BytesInteger(4), 
#-#-#  'ticket_age_add' / BytesInteger(4), 
#-#-#  'ticket_nonce' / Prefixed(BytesInteger(1), GreedyBytes), 
#-#-#  'ticket' / Prefixed(BytesInteger(2), GreedyBytes),
#-#-#  'extensions' / Prefixed(BytesInteger(2), GreedyRange(Extension))
#-#-#)
#-#-#
#-#-#
#-#-### end_of_early_data
#-#-#EndOfEarlyData = Struct ()
#-#-#
#-#-### encrypted extensions
#-#-#EncryptedExtensions = Struct(
#-#-#  '_name' / Computed('EncryptedExtensions'),
#-#-#  'extensions' / Prefixed(BytesInteger(2), GreedyRange(Extension))
#-#-#)
#-#-#
#-#-#
#-#-### Certificate 
#-#-#CertificateType = Enum( BytesInteger(1),
#-#-#  X509 = 0,
#-#-#  RawPublicKey = 2
#-#-#)
#-#-#
#-#-#CertificateEntry = Struct(
#-#-#   
#-#-##   '_certificate_type' / Computed(this._._certificate_type),
#-#-##   'cert' / Switch( this._certificate_type, {
#-#-##    'RawPublicKey': Prefixed( BytesInteger(3), GreedyBytes),
#-#-##    'X509': Prefixed(BytesInteger(3), GreedyBytes)
#-#-##    }
#-#-##  ),
#-#-#  'cert' / Prefixed(BytesInteger(3), GreedyBytes),
#-#-#  'extensions' / Prefixed( BytesInteger(2), GreedyRange(Extension))
#-#-#)
#-#-#
#-#-#Certificate = Struct(
#-#-#  '_name' / Computed('Certificate'),
#-#-##  '_certificate_type' / Computed(this._._certificate_type),
#-#-#  'certificate_request_context' / Prefixed( BytesInteger(1), GreedyBytes),
#-#-#  'certificate_list' / Prefixed(BytesInteger(3), GreedyRange(CertificateEntry))
#-#-#)

## Compressed Certificate
#-#-#CertificateCompressionAlgorithm = Enum ( BytesInteger(2),
#-#-#  zlib = 1,
#-#-#  brotli = 2,
#-#-#  zstd = 3
#-#-#)
#-#-#
#-#-#CompressedCertificate = Struct( 
#-#-#  '_name' / Computed('CompressedCertificate'),
#-#-#  'algorithm' / CertificateCompressionAlgorithm, 
#-#-#  'uncompressed_length' / BytesInteger(3),
#-#-#  'compressed_certificate_message' / Prefixed(BytesInteger(3), GreedyBytes)
#-#-#)
#-#-#
#-#-#
#-#-### CertificateRequest
#-#-#CertificateRequest = Struct(
#-#-#  '_name' / Computed('CertificateRequest'),
#-#-#  'certificate_request_context' / Prefixed( BytesInteger(1), GreedyBytes),
#-#-#  'extensions' / Prefixed(BytesInteger(2), GreedyRange(Extension))
#-#-#)
#-#-#
#-#-### CErtificateVErify
#-#-#CertificateVerify = Struct(
#-#-#  '_name' / Computed('CertificateVerify'),
#-#-#  'algorithm' / SignatureScheme,
#-#-#  'signature' / Prefixed( BytesInteger(2), GreedyBytes)
#-#-#)
#-#-#
#-#-### Finished
#-#-#
#-#-#Finished = Struct(
#-#-#  '_name' / Computed('Finished'),
#-#-###  '_cipher' / Computed(this._._cipher),
#-#-###  '_verify_data_length' / Switch(this._cipher,  
#-#-###    { 'TLS_AES_128_GCM_SHA256' : Const(32, Int8ul),
#-#-###      'TLS_AES_256_GCM_SHA384' : Const(48, Int8ul),
#-#-###      'TLS_CHACHA20_POLY1305_SHA256' : Const(32, Int8ul),
#-#-###      'TLS_AES_128_CCM_SHA256' : Const(32, Int8ul),
#-#-###      'TLS_AES_128_CCM_8_SHA256' : Const(32, Int8ul)
#-#-###    }, default=Rebuild( Byte, len_( this.verify_data ) ) ),
#-#-#  'verify_data' / Select( Bytes( 32 ), Bytes( 48 ) )
#-#-##  'verify_data' / Bytes( this._verify_data_length )  
#-#-#)

##Finished = Struct(
##  '_name' / Computed('Finished'),
##  '_cipher' / Computed(this._._cipher),
##  'verify_data' / Switch(this._cipher,  
##    { 'TLS_AES_128_GCM_SHA256' : Bytes(32),
##      'TLS_AES_256_GCM_SHA384' : Bytes(48),
##      'TLS_CHACHA20_POLY1305_SHA256' : Bytes(32),
##      'TLS_AES_128_CCM_SHA256' : Bytes(32),
##      'TLS_AES_128_CCM_8_SHA256' : Bytes(32)
##    }
##  )
##)

#-#-### KeyUpdate
#-#-#KeyUpdateRequest = Enum( BytesInteger(1), 
#-#-#  update_not_requested = 0, 
#-#-#  update_requested = 1
#-#-#)
#-#-#
#-#-#KeyUpdate = Struct(
#-#-#  '_name' / Computed('KeyUpdate'),
#-#-#  'request_update' / KeyUpdateRequest
#-#-#)
#-#-#
#-#-### Hanshake Message 
#-#-#HandshakeType = Enum( BytesInteger(1), 
#-#-#  client_hello = 1,
#-#-#  server_hello = 2,
#-#-#  new_session_ticket = 4,
#-#-#  end_of_early_data = 5,
#-#-#  encrypted_extensions = 8,
#-#-#  certificate = 11,
#-#-#  certificate_request = 13,
#-#-#  certificate_verify = 15,
#-#-#  finished = 20,
#-#-#  key_update = 24,
#-#-#  message_hash = 254,
#-#-#)
#-#-#
#-#-### certificate_type and cipher can be passed as arguments to the 
#-#-### Handshake structure. These arguments are then passed down to the 
#-#-### downstreamed containers until the final structure. 
#-#-#Handshake = Struct(
#-#-#  '_name' / Computed('Handshake'),
#-#-#  'msg_type' / HandshakeType,
#-#-##  '_certificate_type' / If( this.msg_type == 'certificate', Computed(this._._certificate_type)),   
#-#-###  '_certificate_type' / If( this.msg_type == 'certificate', Computed('X509') ),   
#-#-###  '_cipher' / If( this.msg_type == 'finished', Computed(this._._cipher)),   
#-#-#  'data' / Prefixed( BytesInteger(3), Switch(this.msg_type,
#-#-#    { 'client_hello' : ClientHello,
#-#-#      'server_hello' : ServerHello,
#-#-#      'end_of_early_data' : EndOfEarlyData,
#-#-#      'encrypted_extensions' :  EncryptedExtensions,
#-#-#      'certificate_request' : CertificateRequest,
#-#-#      'certificate': Certificate,
#-#-#      'certificate_verify' : CertificateVerify,
#-#-#      'finished' : Finished,
#-#-#      'new_session_ticket' : NewSessionTicket,
#-#-#      'key_update' : KeyUpdate 
#-#-#    })
#-#-#  )
#-#-#)
#-#-#
#-#-#HandshakeStream = Struct( 
#-#-#  GreedyRange( Handshake )
#-#-#)
#-#-#
#-#-### Designation of specific handshake messages
#-#-### These structures have been designed to designate Sequences. 
#-#-### As Sequences are nesting structures and initializing the 
#-#-### context in the Sequence has not been performed, the parameters
#-#-### associated to the context are read one level above. 
#-#-### Unless there is such need, it is recommended to use the Handshake 
#-#-### and other associated message specific structures
#-#-#
## includes special LURK contructions
HSClientHello = Struct( 
  '_name' / Computed('HSClientHello'),
  'msg_type' / Const('client_hello', tls.HandshakeType),
  'data' / ClientHello
)

### only for LURK
###HSPartialClientHello = Struct( 
###  '_name' / Computed('HSPartialClientHello'),
###  'msg_type' / Const('client_hello', tls.HandshakeType),
###  'data' / PartialClientHello
###)

HSServerHello = Struct(
  '_name' / Computed('HSServerHello'),
  'msg_type' / Const('server_hello', tls.HandshakeType),
  'data' / ServerHello
)
#-#-#
#-#-#HSEncryptedExtensions = Struct( 
#-#-#  '_name' / Computed('HSEncryptedExtensions'),
#-#-#  'msg_type' / Const('encrypted_extensions', HandshakeType), 
#-#-#  'data' / EncryptedExtensions 
#-#-#)
#-#-#
#-#-#HSCertificateRequest = Struct(
#-#-#  '_name' / Computed('HSCertificateRequest'),
#-#-#  'msg_type' / Const('certificate_request', HandshakeType), 
#-#-#  'data' / CertificateRequest
#-#-#)
#-#-#
#-#-#HSCertificate = Struct( 
#-#-##  '_certificate_type' / Computed(this._._._certificate_type),
#-#-#  '_name' / Computed('HSCertificate'),
#-#-#  'msg_type' / Const('certificate', HandshakeType), 
#-#-#  'data' / Certificate
#-#-#)
#-#-#
#-#-#HSCertificateVerify = Struct(
#-#-#  '_name' / Computed('HSCertificateVerify'),
#-#-#  'msg_type' / Const('certificate_verify', HandshakeType), 
#-#-#  'data' / CertificateVerify
#-#-#)
#-#-#
#-#-#HSFinished = Struct(
#-#-#  '_name' / Computed('HSFinished'),
#-#-#  'msg_type' / Const('finished', HandshakeType), 
#-#-#  'data' / Finished 
#-#-#)
#-#-#
#-#-#HSEndOfEarlyData = Struct( 
#-#-#  '_name' / Computed('HSEndOfEarlyData'),
#-#-#  'msg_type' / Const('end_of_early_data', HandshakeType),
#-#-#  'data' / EndOfEarlyData
#-#-#)
#-#-#
#-#-#
#-#-##### Encrypted messages
#-#-#ContentType = Enum( BytesInteger(1), 
#-#-#  invalid = 0,
#-#-#  change_cipher_spec = 20, 
#-#-#  alert = 21,
#-#-#  handshake = 22, 
#-#-#  application_data = 23
#-#-#)
#-#-#
#-#-#TLSPlaintext = Struct(
#-#-#  'type' / ContentType, 
#-#-#  'legacy_record_version' /  Const( b'\x03\x03' ),
#-#-#  'fragment' / Prefixed( BytesInteger(2), Switch( this.type, 
#-#-#     { 'invalid' : GreedyBytes, 
#-#-#       'change_cipher_spec' : GreedyBytes, 
#-#-#       'alert' : GreedyBytes, 
#-#-#       'handshake' : Handshake, 
#-#-#       'application_data' : GreedyBytes } ) )
#-#-#)
#-#-#
#-#-#TLSCiphertext = Struct(
#-#-#  'opaque_type' / Const( ContentType.build( 'application_data' )), 
#-#-#  'legacy_record_version' / Const( b'\x03\x03' ),
#-#-#  'length' / BytesInteger(2), 
#-#-#  'encrypted_reccord' / Bytes( this.length )
#-#-#)
#-#-#
#-#-#
#-#-#TLSInnerPlaintext = Struct(
#-#-#  'content' / GreedyBytes,
#-#-#  'type' / ContentType, 
#-#-#  'zeros' / Array( this._._length_of_padding, Const(b'\x00') ) 
#-#-#)

HkdfLabel = Struct( 
  'length' / BytesInteger( 2 ), 
  'label' / Prefixed(BytesInteger(1), GreedyBytes ),
  'context' / Prefixed(BytesInteger(1), GreedyBytes )
)



#import binascii

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey, X448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
#, Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey
#, Ed448PublicKey
#from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
#, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import \
  SECP256R1, SECP384R1, SECP521R1, ECDSA, \
  EllipticCurvePublicNumbers, ECDH, EllipticCurvePrivateKey
#, EllipticCurvePublicKey
from cryptography.hazmat.primitives.hashes import SHA256, SHA384, SHA512
#Hash,
from cryptography.hazmat.primitives import serialization 
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
# load_der_private_key, load_pem_private_key, NoEncryption, \
#PrivateFormat,
#from cryptography import x509
#from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives.asymmetric import padding
#from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305, AESGCM, AESCCM
from cryptography.hazmat.primitives.kdf.hkdf import HKDFExpand
#HKDF,

from pylurk.lurk.lurk_lurk import LURKError, ImplementationError
#, ConfigurationError
import pylurk.tls13.struct_tls13 as lurk
import pylurk.debug
import pytls13.struct_tls13 as tls

def hash_sanity_check( tls_hash ):
  """ enables to provide tls_hash as a string

  This is especially useful when psk_metadata structure or tickets
  are carried as structure as opposed to objects.
  """
  if isinstance( tls_hash, ( SHA256, SHA384, SHA512 ) ): # or \
#     isinstance( tls_hash, SHA384() ) or \
#     isinstance( tls_hash, SHA512() ):
    return tls_hash

  if isinstance( tls_hash, str ) :
    if tls_hash == 'sha256' :
      tls_hash = SHA256()
    elif tls_hash == 'sha384' :
      tls_hash = SHA384()
    elif tls_hash == 'sha384' :
      tls_hash = SHA512()
    else :
      raise ImplementationError( f"unknown tls_hash {tls_hash}" )
  else :
    raise ImplementationError( f"unknown tls_hash {tls_hash}" )
  return tls_hash


### we may need to change this to SigScheme
# class SigAlgo
class SigScheme:
  def __init__( self, name:str) :
    """ used to pars ethe signature scheme

    The function also extract the hash for cipher suites
    """
    self.name = name
    self.algo = self.get_algo()
    self.hash = self.get_hash()
    self.pad = self.get_pad()
    self.curve = self.get_curve()


  def get_algo( self ):
    return self.name.split('_')[0]

  def get_hash( self ):
    ## returns None in case of ed25519 or ed448 sig scheme
    ## TLS cipher suite are expected to pass.
    ## we use hashlib to remain compatible with key_schedule
    if self.algo in [ 'ed25519', 'ed448' ]:
      return None
    hash_algo = self.name.split( '_' )[-1].lower()
    try:
      return hash_sanity_check( hash_algo )
    except :
##    if hash_algo == 'sha256':
##      h = SHA256()
###      h = hashlib.sha256
##    elif hash_algo == 'sha384':
##      h = SHA384()
###      h = hashlib.sha384
##    elif hash_algo == 'sha512':
##      h = SHA512()
###      h = hashlib.sha512
##    else:
      raise LURKError( 'invalid_signature_scheme', f"{hash_algo} is not implemented" )
##    return h

  def get_curve( self ): # -> Union[ SECP256R1, SECP384R1, SECP521R1 ] :
    if self.algo != 'ecdsa':
      return None
    curve_name = self.name.split( '_' )[1]
    if  curve_name == 'secp256r1':
      curve = SECP256R1()
    elif curve_name == 'secp384r1':
      curve = SECP384R1()
    elif curve_name == 'secp521r1':
      curve = SECP521R1()
    else:
      raise LURKError( 'invalid_signature_scheme', f"{curve_name} is not implemented" )
    return curve

  def get_pad( self ):
    if self.algo != 'rsa':
      return None
    pad_name = self.name.split( '_' )[1]
    if pad_name == 'pkcs1':
      pad = padding.PKCS1v15()
    elif pad_name == 'pss':
#      pad = padding.PSS(
#        mgf=padding.MGF1(self.hash),
#       salt_length=padding.PSS.MAX_LENGTH)
      pad = padding.PSS(
        mgf=padding.MGF1(self.hash),
       salt_length=self.hash.digest_size )
    else:
      raise LURKError( 'invalid_signature_scheme', f"{pad_name} is not implemented" )
    return pad

  def matches( self, key):
    if ( self.algo == 'ed25519' and not \
         isinstance( key, Ed25519PrivateKey ) ) or\
       ( self.algo == 'ed448' and not \
         isinstance( key, Ed448PrivateKey ) ) or\
       ( self.algo == 'ecdsa' and not \
          isinstance( key, EllipticCurvePrivateKey ) ) or\
       ( self.algo == 'rsa' and not \
          isinstance( key, RSAPrivateKey ) ):
      raise LURKError( 'invalid_signature_scheme', \
              f"{self.name}, {type(key)} ,"\
              f"incompatible private key and signature algorithm" )
    if isinstance( key, EllipticCurvePrivateKey ):
      if isinstance( key.curve, type( self.curve ) ) is False:
        raise LURKError( 'invalid_signature_scheme', \
              f"{self.name}, {self.curve}, {key.curve} ,"\
              f"incompatible curve and signature algorithm" )



class CipherSuite:
  def __init__( self, name:str, secret=None ) :
    """ Handle the cipher suite string
    """
    self.name = name
    self.hash = self.get_hash()
    self.tag_len = self.tag_length( )
    self.key_len = self.key_length( )
    self.nonce_len = self.nonce_length( )

    if secret is not None:
      self.traffic_key( secret )


  def get_hash( self ):
    return SigScheme( self.name ).get_hash( )

  def nonce_length( self ):
    return 12

  def tag_length( self ):
    if self.name == 'TLS_CHACHA20_POLY1305_SHA256':
      tag_length = 16
    elif self.name == 'TLS_AES_128_GCM_SHA256':
      tag_length = 16
    elif self.name == 'TLS_AES_256_GCM_SHA384':
      tag_length = 16
    elif self.name == 'TLS_AES_128_CCM_SHA256':
      tag_length = 16
    elif self.name == 'TLS_AES_128_CCM_8_SHA256':
      tag_length = 8
    else:
      raise LURKError( 'invalid_cipher_suite', f"{self.name} is not implemented" )
    return tag_length

  def key_length( self ):
    if self.name == 'TLS_CHACHA20_POLY1305_SHA256':
      key_length = 32
    elif self.name == 'TLS_AES_128_GCM_SHA256':
      key_length = 16
    elif self.name == 'TLS_AES_256_GCM_SHA384':
      key_length = 32
    elif self.name == 'TLS_AES_128_CCM_SHA256':
      key_length = 16
    elif self.name == 'TLS_AES_128_CCM_8_SHA256':
      key_length = 16
    else:
      raise LURKError( 'invalid_cipher_suite', f"{self.name} is not implemented" )
    return key_length

  def compute_nonce( self, sequence_number, iv ):
    sn = (b'\x00' * 4) + int(sequence_number).to_bytes( 8, byteorder='big' )
    xor = bytearray()
    for sn, iv in zip( sn, iv ):
      xor.append( sn ^ iv )
    return xor
#      nonce = b''
#      for i in range( self.nonce_len ):
#        nonce += sn[ i ] ^ iv[ i ]
#      return nonce
#      formatted_num = (b"\x00" * 4) + struct.pack(">q", num)
#    return bytes([i ^ j for i, j in zip(iv, formatted_num)])

  def hkdf_expand_label( self, secret,\
                       label, \
                       context, \
                       length, \
                       backend=default_backend() ): #, \
    hkdf_label = lurk.HkdfLabel.build( \
      { 'length' : length,
        'label' : b"tls13 " + label, \
        'context' : context } )
    return HKDFExpand( algorithm=self.hash,\
                         length=length,\
                         info=hkdf_label,\
                         backend=backend ).derive( secret )

  def traffic_key( self, secret ):
#    ks = pylurk.tls13.lurk_tls13.KeyScheduler( tls_hash=self.hash )
    self.write_key = self.hkdf_expand_label( secret=secret, label=b'key', \
            context=b'', length=self.key_len )
    self.write_iv = self.hkdf_expand_label( secret=secret, label=b'iv', \
            context=b'', length=self.nonce_len )
    self.sequence_number = 0


  def next_generation_application_traffic_secret( self, secret ):
    return self.hkdf_expand_label( secret=secret, label=b'traffic upd', \
             context=b'', length=self.hash.digest_size )

  def decrypt_old( self, msg, debug=False ):
    """ decrypt msg and return a plain text structure (equivalent)

    """
    additional_data = b'\x17\x03\x03' + len( msg ).to_bytes( 2, byteorder='big' )
    nonce = self.compute_nonce( self.sequence_number, self.write_iv )
    pylurk.debug.print_bin( "fragment (encrypted)",  msg  )
    pylurk.debug.print_bin( "write_key", self.write_key )
    pylurk.debug.print_bin( "write_iv", self.write_iv )
    pylurk.debug.print_bin( "nonce", nonce )
    pylurk.debug.print_bin( "additional_data", additional_data )
    pylurk.debug.print_val( 'sequence_number', self.sequence_number )
    if 'GCM' in self.name:
      cipher = AESGCM( self.write_key )
    elif 'CCM' in self.name :
      cipher = AESCCM( self.write_key, tag_length=self.tag_len )
    elif self.name == 'TLS_CHACHA20_POLY1305_SHA256':
      cipher = ChaCha20Poly1305( self.write_key )
    else:
      raise LURKError( 'invalid_cipher_suite', f"{self.name} is not implemented" )
    clear_text = cipher.decrypt( nonce, msg, additional_data )
    ## this probably can be handled by construct itself
    length_of_padding = 0
    for i in range( len( clear_text ) ):
      if clear_text[ -1 - i ] != b'\x00':
        break
      else:
        length_of_padding += 1
    type_byte = ( clear_text[ -1 - length_of_padding ]).to_bytes( 1, byteorder='big' )
    ct_type = tls.ContentType.parse( type_byte )
    pylurk.debug.print_bin( f"fragment (decrypted) [type {ct_type}]",  clear_text )
    if ct_type in [ 'application_data', 'handshake' ] :
      clear_text_msg_len = len( clear_text ) - 1 - length_of_padding
      clear_text_struct = tls.TLSInnerPlaintext.parse( clear_text, \
              type=ct_type, length_of_padding=length_of_padding, \
              clear_text_msg_len=clear_text_msg_len )
    else:
      clear_text_struct = tls.TLSInnerPlaintext.parse( clear_text, \
              type=ct_type, length_of_padding=length_of_padding )
    self.sequence_number += 1
    if debug is True:
      return clear_text_struct, clear_text
    return  { 'type' : ct_type, 'content' : clear_text_struct[ 'content' ] }


  def decrypt( self, msg:bytes, debug=None ) -> dict:
    """ decrypt msg and return a plain text structure (equivalent)

    """
    additional_data = b'\x17\x03\x03' + len( msg ).to_bytes( 2, byteorder='big' )
    nonce = self.compute_nonce( self.sequence_number, self.write_iv )
    if debug is not None :
      pylurk.debug.print_bin( "fragment (encrypted)",  msg  )
      pylurk.debug.print_bin( "write_key", self.write_key )
      pylurk.debug.print_bin( "write_iv", self.write_iv )
      pylurk.debug.print_bin( "nonce", nonce )
      pylurk.debug.print_bin( "additional_data", additional_data )
      pylurk.debug.print_val( 'sequence_number', self.sequence_number )
    if 'GCM' in self.name:
      cipher = AESGCM( self.write_key )
    elif 'CCM' in self.name :
      cipher = AESCCM( self.write_key, tag_length=self.tag_len )
    elif self.name == 'TLS_CHACHA20_POLY1305_SHA256':
      cipher = ChaCha20Poly1305( self.write_key )
    else:
      raise LURKError( 'invalid_cipher_suite', f"{self.name} is not implemented" )
    clear_text = cipher.decrypt( nonce, msg, additional_data )
    ## this probably can be handled by construct itself
    length_of_padding = 0
    for i in range( len( clear_text ) ):
      if clear_text[ -1 - i ] != b'\x00':
        break
      else:
        length_of_padding += 1
    type_byte = ( clear_text[ -1 - length_of_padding ]).to_bytes( 1, byteorder='big' )
    ct_type = tls.ContentType.parse( type_byte )
    if ct_type in [ 'application_data', 'handshake' ] :
      clear_text_msg_len = len( clear_text ) - 1 - length_of_padding
      clear_text_struct = tls.FragmentTLSInnerPlaintext.parse( \
              clear_text, type=ct_type, \
              length_of_padding=length_of_padding, \
              clear_text_msg_len=clear_text_msg_len )
    else:
      clear_text_struct = tls.TLSInnerPlaintext.parse( \
              clear_text, type=ct_type, \
              length_of_padding=length_of_padding )
    self.sequence_number += 1
    return  { 'type' : ct_type, \
              'content' : clear_text_struct[ 'content' ],\
              'zeros' : b'\x00' * length_of_padding }


  def encrypt( self, clear_text_msg, content_type,\
               length_of_padding=0, debug=None ):
    """ builds and encrypts the clear_text_msg as inner_text message """

    ## building the inner message
    zeros = b'\x00' * length_of_padding
    inner_plain_text = { \
      'content' : clear_text_msg,
      'type' : content_type,
      'zeros' : zeros }
    if content_type == 'application_data' :
      clear_text_record_bytes = tls.TLSInnerPlaintext.build( \
              inner_plain_text, type=content_type, \
              length_of_padding=length_of_padding, \
              clear_text_msg_len=len(clear_text_msg) )
    else: # handshake
      clear_text_record_bytes = tls.TLSInnerPlaintext.build( \
              inner_plain_text, type=content_type, \
              length_of_padding=length_of_padding)

    additional_data = b'\x17\x03\x03' + \
            int( len( clear_text_record_bytes ) +\
            self.tag_len ).to_bytes( 2, byteorder='big' )
    nonce = self.compute_nonce( self.sequence_number, self.write_iv )
    if 'GCM' in self.name:
      cipher = AESGCM( self.write_key )
    elif 'CCM' in self.name :
      cipher = AESCCM( self.write_key, tag_length=self.tag_len )
    elif self.name == 'TLS_CHACHA20_POLY1305_SHA256':
      cipher = ChaCha20Poly1305( self.write_key )
    else:
      raise LURKError( 'invalid_cipher_suite', f"{self.name} is not implemented" )
    encrypted_reccord =  cipher.encrypt( nonce, clear_text_record_bytes, additional_data )
    if debug is True :
      pylurk.debug.print_bin( "write_key", self.write_key )
      pylurk.debug.print_bin( "write_iv", self.write_iv )
      print( f"  - sequence_number : {self.sequence_number}" )
      pylurk.debug.print_bin( "nonce", nonce )
      pylurk.debug.print_bin( "additional_data", additional_data )

    self.sequence_number += 1
    return encrypted_reccord

  def debug( self, debug, description="" ):
    debug.handle_bin( f"{description}_write_key", self.write_key )
    debug.handle_bin( f"{description}_write_iv", self.write_iv )


class ECDHEKey():

  def __init__( self ):
    self.group = None
    self.private = None
    self.public = None

  def generate( self, group ):
    """ generates a brand ne ECDHE key pair for a given group """
    self.group = group
    if group in [ 'secp256r1', 'secp384r1', 'secp521r1' ]:
      if group ==  'secp256r1':
        curve =  SECP256R1()
      elif group ==  'secp384r1':
        curve = SECP384R1()
      elif group ==  'secp521r1':
        curve = SECP521R1()
#      private_key = ec.generate_private_key( curve, default_backend())
      private_key = ec.generate_private_key( curve )
#      public_key = private_key.public_key()
    elif group  in [ 'x25519', 'x448' ]:
      if group == 'x25519':
        private_key = X25519PrivateKey.generate()
      elif group == 'x448':
        private_key = X448PrivateKey.generate()
    self.private = private_key
    self.public = private_key.public_key()

  def ks_entry( self ) -> dict:
    """ returns the ks_entry (TLS) associated to the ECDHE key

    Note that only the public part is in the ks entry
    """
    if self.group in [ 'secp256r1', 'secp384r1', 'secp521r1' ]:
      public_numbers = self.public.public_numbers()
      key_exchange = { 'legacy_form' : 4,
                       'x' : public_numbers.x,
                       'y' : public_numbers.y }
    elif self.group  in [ 'x25519', 'x448' ]:
      key_exchange = self.public.public_bytes(
        encoding=Encoding.Raw, format=PublicFormat.Raw)
    return { 'group' : self.group,
             'key_exchange' : key_exchange }

  def generate_from_pem( self, pem_bytes:bytes ):
## private_key = serialization.load_pem_private_key(\
## b'-----BEGIN PRIVATE KEY-----\n\
## MC4CAQAwBQYDK2VuBCIEICAhIiMkJSYnKCkqKywtLi8wMTIzNDU2Nzg5Ojs8PT4/\n\
## -----END PRIVATE KEY-----', None)
    self.private = serialization.load_pem_private_key( pem_bytes, None)
    self.public = self.private.public_key()

  def pkcs8( self ):
    if self.private is not None:
      private_bytes = self.private.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption() )
    else:
      private_bytes = b''
    return private_bytes

  def generate_from_ks_entry( self, ks_entry:dict ):
    """ generates the ECHDE key from a ks_entry structure

    Note that only the public part is in the ks entry.

    .. code-block:: python

       ks_entry is a dictionary { 'group' : 'x25519', 
                                  'key_exchange' : kx_bytes }

    """

    self.group = ks_entry[ 'group' ]
    key_exchange = ks_entry[ 'key_exchange' ]
##    if group not in self.conf[ 'authorized_ecdhe_group' ]:
##      raise LURKError( 'invalid_ephemeral', f"unsupported {self.group}" )
    if self.group in [ 'secp256r1', 'secp384r1', 'secp521r1' ]:
      if self.group ==  'secp256r1':
        curve =  SECP256R1()
      elif self.group ==  'secp384r1':
        curve = SECP384R1()
      elif self.group ==  'secp521r1':
        curve = SECP521R1()
      else:
        raise LURKError( 'invalid_ephemeral', f"unknown group {self.group}" )

      public_number = EllipticCurvePublicNumbers( key_exchange[ 'x' ],\
                     key_exchange[ 'y' ], curve )
      self.public = public_number.public_key()
    elif self.group  in [ 'x25519', 'x448' ]:
      if self.group == 'x25519':
        self.public = X25519PublicKey.from_public_bytes( key_exchange )
      elif self.group == 'x448':
        self.public = X448PublicKey.from_public_bytes( key_exchange )
    else:
      raise LURKError( 'invalid_ephemeral', f"unknown group {self.group}" )

  def shared_secret( self, ecdhe_key ):
    if self.private is None:
      private_key = ecdhe_key.private
      public_key = self.public
    else:
      private_key = self.private
      public_key = ecdhe_key.public
    if self.group in [ 'secp256r1', 'secp384r1', 'secp521r1' ]:
      print( f"private key {type(private_key)}" )
      print( f"public key {type(public_key)}" )
      print( f"ECDHE {type(ECDH())}" )
      shared_secret = private_key.exchange( ECDH(), public_key )
    elif self.group  in [ 'x25519', 'x448' ]:
      shared_secret = private_key.exchange( public_key )
    else:
      raise LURKError( 'invalid_ephemeral', f"Unexpected group {self.group}" )
    return shared_secret



